from collections import defaultdict
import datetime
import json
import mock
import requests
import unittest

import uareach as ua
from uareach import common
from uareach.templates import Template
from template_builders import build_apple_loyalty, build_google_loyalty


class TemplateApiTest(unittest.TestCase):

    def setUp(self):
        self.client = ua.Reach('fake', 'creds')
        super(TemplateApiTest, self).setUp()

    @mock.patch.object(ua.Reach, '_request')
    def test_delete(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        mock_request.return_value = response

        reach_response = ua.delete_template(self.client, template_id=12345)
        mock_request.assert_called_with(
            'DELETE',
            None,
            common.TEMPLATE_BASE_URL.format(12345),
            None,
            1.2,
            None
        )

        self.assertEqual(reach_response, True)

    @mock.patch.object(ua.Reach, '_request')
    def test_add_template_locations(self, mock_request):
        location = {
            "longitude": -122.374,
            "latitude": 37.618,
            "relevantText": "Hello loc0",
            "streetAddress1": "address line #1",
            "streetAddress2": "address line #2",
            "city": "Palo Alto",
            "region": "CA",
            "regionCode": "94404",
            "country": "US"
        }
        body = json.dumps({
            'locations': [location]
        })
        response = requests.Response()
        response._content = json.dumps([
            {
                'value': location,
                'locationId': 231,
                'fieldId': 312
            }
        ])
        mock_request.return_value = response

        reach_response = ua.add_template_locations(
            self.client, [location], template_id=12345
        )
        mock_request.assert_called_with(
            'POST',
            body,
            ua.common.TEMPLATE_ADD_LOCATION_URL.format(12345),
            'application/json',
            1.2,
            None
        )
        self.assertEqual(reach_response[0]['value'], location)
        self.assertEqual(reach_response[0]['locationId'], 231)
        self.assertEqual(reach_response[0]['fieldId'], 312)

    @mock.patch.object(ua.Reach, '_request')
    def test_remove_template_location(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        mock_request.return_value = response

        reach_response = ua.delete_template_location(
            self.client, 'location123', template_id=12345
        )
        mock_request.assert_called_with(
            'DELETE',
            None,
            common.TEMPLATE_REMOVE_LOCATION_URL.format(12345, 'location123'),
            None,
            1.2,
            None
        )
        self.assertEqual(reach_response, True)

    @mock.patch.object(ua.Reach, '_request')
    def test_duplicate_template(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        response._content = json.dumps(
            {'templateId': 54321}
        )
        mock_request.return_value = response
        reach_response = ua.duplicate_template(self.client, template_id=12345)
        mock_request.assert_called_with(
            'POST',
            None,
            ua.common.TEMPLATE_DUPLICATE_URL.format(12345),
            None,
            1.2,
            None
        )
        self.assertEqual(reach_response['templateId'], 54321)

    @mock.patch.object(ua.Reach, '_request')
    def test_get_apple_template(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        template_json = json.dumps({
            'fieldsModel': {
                'fields': {
                    'Points': {
                        'fieldType': 'primary',
                        "numberStyle": 'numberStyleDecimal',
                        'value': 33.0
                    }
                },
                'headers': {
                    'barcode_encoding': {
                        'fieldType': 'barcode',
                        'formatType': 'String',
                        'value': 'iso-8859-1'
                    },
                }
            },
            'templateHeader': {
                'description': 'Hello',
                'name': 'Test',
                'type': 'Loyalty',
                'vendor': 'Apple',
                'vendorId': 1
            }
        })
        response._content = template_json
        mock_request.return_value = response
        ua.get_template(self.client, template_id=12345)
        mock_request.assert_called_with(
            'GET',
            None,
            common.TEMPLATE_BASE_URL.format(12345),
            None,
            1.2,
            None
        )

    @mock.patch.object(ua.Reach, '_request')
    def test_get_google_template(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        template_json = json.dumps({
            'fieldsModel': {
                'infoModuleData': {
                    'backgroundColor': {
                        'formatType': 'String',
                        'fieldType': 'infoModuleData',
                        'value': 339
                    }
                }
            },
            'headers': {
                'barcode_encoding': {
                    'fieldType': 'topLevel',
                    'formatType': 'String',
                    'value': 'iso-8859-1',
                    'label': '',
                },
            },
            'templateHeader': {
                'vendor': 'Google',
                'projectType': "memberCard",
                'projectId': 3456,
                'description': 'Hello',
                'type': 'Loyalty',
                'vendorId': 2,
                'name': 'Test'
            }
        })
        response._content = template_json
        mock_request.return_value = response
        ua.get_template(self.client, template_id=67890)
        mock_request.assert_called_with(
            'GET',
            None,
            ua.common.TEMPLATE_BASE_URL.format(67890),
            None,
            1.2,
            None
        )

    @mock.patch.object(ua.Reach, '_request')
    def test_listing(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        templates_json = json.dumps({
            'count': 133,
            'templateHeaders': [
                {
                   'vendor': 'Google',
                   'projectType': 'memberCard',
                   'projectId': '10057',
                   'type': 'Loyalty1',
                   'vendorId': '2',
                   'deleted': 'False',
                   'id': '23595',
                   'updatedAt': '2013-07-01T18:28:54.000Z',
                   'description': 'description',
                   'createdAt': '2013-07-01T18:28:54.000Z',
                   'name': 'New Reach Template',
                   'disabled': 'False'
                },
                {
                   'vendor': 'Apple',
                   'projectType': 'memberCard',
                   'projectId': '10057',
                   'type': 'Store Card',
                   'vendorId': '1',
                   'deleted': 'False',
                   'id': '23593',
                   'updatedAt': '2013-07-01T18:28:33.000Z',
                   'description': 'Description',
                   'createdAt': '2013-07-01T18:28:33.000Z',
                   'name': 'Loyalty Card',
                   'disabled': 'False'
                }
            ],
            'pagination': {
                'order': 'id',
                'page': 1,
                'start': 0,
                'direction': 'DESC',
                'pageSize': 10
            }
        })
        response._content = templates_json
        mock_request.return_value = response
        listing_response = ua.TemplateList(self.client)
        self.assertEqual(
            listing_response.next()['id'],
            '23595'
        )
        mock_request.assert_called_with(
            'GET',
            None,
            common.TEMPLATE_BASE_URL.format('headers'),
            None,
            1.2,
            {}
        )


class TemplateTest(unittest.TestCase):

    def setUp(self):
        self.template = Template()
        self.template.add_metadata(name='Test generic template')

    def test_add_fields(self):
        member_name = ua.Field(
            name='Member Name',
            label='Member Name',
            value='First Last',
            fieldType=ua.AppleFieldType.PRIMARY
        )
        merchant_website = ua.Field(
            name='Merchant Website',
            label='Merchant Website',
            value='http://test.com',
            fieldType=ua.AppleFieldType.BACK
        )
        self.template.add_fields(member_name, merchant_website)
        self.assertEqual(self.template.fields['Member Name'], member_name)
        self.assertEqual(
            self.template.fields['Merchant Website'], merchant_website
                         )

    def test_remove_fields(self):
        member_name = ua.Field(
            name='Member Name',
            label='Member Name',
            value='First Last',
            fieldType=ua.AppleFieldType.PRIMARY
        )
        merchant_website = ua.Field(
            name='Merchant Website',
            label='Merchant Website',
            value='http://test.com',
            fieldType=ua.AppleFieldType.BACK
        )
        self.template.add_fields(member_name, merchant_website)
        self.assertNotEqual(self.template.fields, defaultdict(dict))
        self.template.remove_fields('Member Name', 'Merchant Website')
        self.assertEqual(self.template.fields, defaultdict(dict))

    def test_set_fields(self):
        member_name = ua.Field(
            name='Member Name',
            label='Member Name',
            value='First Last',
            fieldType=ua.AppleFieldType.PRIMARY
        )
        merchant_website = ua.Field(
            name='Merchant Website',
            label='Merchant Website',
            value='http://test.com',
            fieldType=ua.AppleFieldType.BACK
        )
        self.template.add_fields(member_name, merchant_website)
        self.assertNotEqual(self.template.fields, defaultdict(dict))
        points = ua.Field(
            name='Reward Points',
            label='Reward Points',
            value=1234.0,
            fieldType=ua.AppleFieldType.SECONDARY
        )
        self.template.set_fields(points)
        self.assertEqual(len(self.template.fields), 1)
        self.assertEqual(self.template.fields['Reward Points'], points)

    def test_add_metadata(self):
        self.template.add_metadata(
            description='a description!',
            vendorId=1,
            vendor='Apple'
        )
        self.assertEqual(
            self.template.metadata['description'], 'a description!'
        )
        self.assertEqual(self.template.metadata['vendorId'], 1)
        self.assertEqual(self.template.metadata['vendor'], 'Apple')

    def test_remove_metadata(self):
        self.template.add_metadata(
            description='a description!',
            vendorId=1,
            vendor='Apple'
        )
        self.assertEqual(len(self.template.metadata), 4)
        self.template.remove_metadata('vendorId', 'vendor')
        self.assertEqual(len(self.template.metadata), 2)
        self.assertIsNone(self.template.metadata.get('vendorId', None))
        self.assertIsNone(self.template.metadata.get('vendor', None))

    def test_set_metadata(self):
        self.template.add_metadata(
            description='a description!',
            vendorId=1,
            vendor='Apple'
        )
        self.assertEqual(len(self.template.metadata), 4)
        self.template.set_metadata(
            name='new name!!',
            description='new description',
            vendorId=2,
            vendor='Google'
        )
        self.assertEqual(
            self.template.metadata,
            {
                'name': 'new name!!',
                'description': 'new description',
                'vendorId': 2,
                'vendor': 'Google'
            }
        )

    def test_add_header(self):
        self.template.add_metadata(vendor='Apple')
        self.template.add_header('barcode_value', '123456789')
        self.template.add_header('logo_color', ua.rgb(23, 23, 23))
        self.assertEqual(
            self.template.headers['barcode_value'],
            {'value': '123456789', 'fieldType': 'barcode', 'formatType': 'String'}
        )
        self.assertEqual(
            self.template.headers['logo_color'],
            {
                'value': 'rgb(23,23,23)',
                'fieldType': 'topLevel',
                'formatType': 'String'
            }
        )
        self.assertRaises(
            ValueError,
            lambda: self.template.add_header('non_existent_header', 'blah')
        )

    def test_add_headers(self):
        self.template.add_metadata(vendor='Apple')
        self.template.add_headers(
            barcode_value='123456789',
            logo_color=ua.rgb(23, 23, 23)
        )
        self.assertEqual(
            self.template.headers,
            {
                'barcode_value': {
                    'value': '123456789',
                    'fieldType': 'barcode',
                    'formatType': 'String'
                },
                'logo_color': {
                    'value': 'rgb(23,23,23)',
                    'fieldType': 'topLevel',
                    'formatType': 'String'
                }
            }
        )

    def test_remove_headers(self):
        self.template.add_metadata(vendor='Apple')
        self.template.add_headers(
            barcode_value='123456789',
            logo_color=ua.rgb(23, 23, 23)
        )
        self.assertNotEqual(self.template.headers, {})
        self.template.remove_headers('barcode_value', 'logo_color')
        self.assertEqual(self.template.headers, {})

    def test_set_headers(self):
        self.template.add_metadata(vendor='Apple')
        self.template.add_headers(
            barcode_value='123456789',
            logo_color=ua.rgb(23, 23, 23)
        )
        self.assertNotEqual(self.template.headers, {})
        self.template.set_headers(
            background_color=ua.rgb(0, 147, 201),
            logo_text='logo text',
            thumbnail_image='https://google.com/images/cool_gif_bro.gif'
        )
        self.assertEqual(
            self.template.headers,
            {
                'background_color': {
                    'value': 'rgb(0,147,201)',
                    'fieldType': 'topLevel',
                    'formatType': 'String'
                },
                'logo_text': {
                    'value': 'logo text',
                    'fieldType': 'topLevel',
                    'formatType': 'String'
                },
                'thumbnail_image': {
                    'value': 'https://google.com/images/cool_gif_bro.gif',
                    'fieldType': 'image',
                    'formatType': 'String'
                }
            }
        )

    def test_build_url(self):
        # Lookup template URLs
        get_template_url = Template.build_url(
            common.TEMPLATE_BASE_URL, main_id=1234
        )
        self.assertEqual(
            get_template_url, common.TEMPLATE_BASE_URL.format(1234)
        )
        get_template_url = Template.build_url(
            common.TEMPLATE_BASE_URL, external_id=1234
        )
        self.assertEqual(
            get_template_url, common.TEMPLATE_BASE_URL.format('id/1234')
        )
        # Create template URLs
        create_template_url = Template.build_url(
            common.TEMPLATE_BASE_URL, main_id=1234
        )
        self.assertEqual(
            create_template_url, common.TEMPLATE_BASE_URL.format(1234)
        )
        create_template_url = Template.build_url(
            common.TEMPLATE_BASE_URL, main_id=1234, external_id=5678
        )
        self.assertEqual(
            create_template_url,
            common.TEMPLATE_BASE_URL.format('1234/id/5678')
        )
        # Delete template URLs
        delete_template_url = Template.build_url(
            common.TEMPLATE_BASE_URL, main_id=1234
        )
        self.assertEqual(
            delete_template_url, common.TEMPLATE_BASE_URL.format(1234)
        )
        delete_template_url = Template.build_url(
            common.TEMPLATE_BASE_URL, external_id=1234
                                               )
        self.assertEqual(
            delete_template_url, common.TEMPLATE_BASE_URL.format('id/1234')
        )
        # Duplicate template URLs
        duplicate_url = Template.build_url(
            common.TEMPLATE_DUPLICATE_URL, main_id=1234
        )
        self.assertEqual(
            duplicate_url, common.TEMPLATE_DUPLICATE_URL.format(1234)
        )
        duplicate_url = Template.build_url(
            common.TEMPLATE_DUPLICATE_URL, external_id=1234
        )
        self.assertEqual(
            duplicate_url, common.TEMPLATE_DUPLICATE_URL.format('id/1234')
        )
        # Add template location URLs
        add_location_url = Template.build_url(
            common.TEMPLATE_ADD_LOCATION_URL, main_id=1234
                                            )
        self.assertEqual(
            add_location_url, common.TEMPLATE_ADD_LOCATION_URL.format(1234)
        )
        add_location_url = Template.build_url(
            common.TEMPLATE_ADD_LOCATION_URL, external_id=1234
                                            )
        self.assertEqual(
            add_location_url,
            common.TEMPLATE_ADD_LOCATION_URL.format('id/1234')
        )
        # Remove template location URLs
        rem_location_url = Template.build_url(
            common.TEMPLATE_REMOVE_LOCATION_URL,
            main_id=1234,
            location_id=5678
        )
        self.assertEqual(
            rem_location_url,
            common.TEMPLATE_REMOVE_LOCATION_URL.format(1234, 5678)
        )
        rem_location_url = Template.build_url(
            common.TEMPLATE_REMOVE_LOCATION_URL,
            external_id=1234,
            location_id=5678
        )
        self.assertEqual(
            rem_location_url,
            common.TEMPLATE_REMOVE_LOCATION_URL.format('id/1234', 5678)
        )


class AppleTemplateTest(unittest.TestCase):

    def setUp(self):
        self.template = build_apple_loyalty()
        self.client = ua.Reach('fake', 'creds')
        super(AppleTemplateTest, self).setUp()

    @mock.patch.object(ua.Reach, '_request')
    def test_create_template(self, mock_request):
        response = requests.Response()
        response._content = json.dumps({'templateId': 54123})
        response.status_code = 200
        mock_request.return_value = response

        create_response = self.template.create(self.client, project_id=12345)
        mock_request.assert_called_with(
            'POST',
            json.dumps(self.template._create_payload()),
            common.TEMPLATE_BASE_URL.format(12345),
            'application/json',
            1.2,
            None
        )
        self.assertEqual(create_response['templateId'], 54123)

    @mock.patch.object(ua.Reach, '_request')
    def test_update_template(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        mock_request.return_value = response
        # self.template.add_metadata(id_=12345)

        update_response = self.template.update(self.client, template_id=12345)
        mock_request.assert_called_with(
            'PUT',
            json.dumps(self.template._create_payload()),
            common.TEMPLATE_BASE_URL.format(12345),
            'application/json',
            1.2,
            None
        )
        self.assertEqual(update_response, True)

    def test_add_header(self):
        self.assertEqual(
            self.template.headers['barcode_type']['value'],
            ua.BarcodeType.convert_to_apple(ua.BarcodeType.PDF_417)
        )

        # Check barcode validation
        self.assertRaises(
            ValueError,
            lambda: self.template.add_headers(barcode_type=ua.BarcodeType.EAN_13)
        )

    # Add another test outside of the test_update_template?
    def test_add_metadata(self):
        pass

    def test_add_beacon(self):
        self.template.add_beacon(
            'f7c7066e-ef28-4679-87c5-df3686a3dddb',
            relevant_text='Hello new beacon'
        )
        self.assertEqual(len(self.template.beacons), 2)
        for beacon in self.template.beacons:
            if beacon['uuid'] == 'f7c7066e-ef28-4679-87c5-df3686a3dddb':
                self.assertEqual(beacon['relevantText'], 'Hello new beacon')

    def test_remove_beacon(self):
        self.template.remove_beacon('3526dee6-4ea8-11e6-beb8-9e71128cae77')
        self.assertEqual(self.template.beacons, [])


class GoogleTemplateTest(unittest.TestCase):

    def setUp(self):
        self.template = build_google_loyalty()
        self.client = ua.Reach('fake', 'creds')
        super(GoogleTemplateTest, self).setUp()

    @mock.patch.object(ua.Reach, '_request')
    def test_create_template(self, mock_request):
        response = requests.Response()
        response._content = json.dumps({'templateId': 54321})
        response.status_code = 200
        mock_request.return_value = response

        create_response = self.template.create(self.client, project_id=12345)
        mock_request.assert_called_with(
            'POST',
            json.dumps(self.template._create_payload()),
            common.TEMPLATE_BASE_URL.format(12345),
            'application/json',
            1.2,
            None
        )
        self.assertEqual(create_response['templateId'], 54321)

    @mock.patch.object(ua.Reach, '_request')
    def test_update_template(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        mock_request.return_value = response
        # self.template.add_metadata(id_=12345)

        update_response = self.template.update(self.client, template_id=54321)
        mock_request.assert_called_with(
            'PUT',
            json.dumps(self.template._create_payload()),
            common.TEMPLATE_BASE_URL.format(54321),
            'application/json',
            1.2,
            None
        )
        self.assertEqual(update_response, True)

        # Add another test outside of the test_update_template?
        def test_add_metadata(self):
            pass

    def test_add_header(self):
        self.assertEqual(
            self.template.headers['barcode_type']['value'],
            ua.BarcodeType.PDF_417
        )

        self.template.add_headers(
            barcode_type=ua.BarcodeType.AZTEC
        )
        self.assertEqual(
            self.template.headers['barcode_type']['value'],
            ua.BarcodeType.AZTEC
        )
        self.assertRaises(
            ValueError,
            lambda: self.template.add_headers(logo_color=ua.rgb(12, 12, 12))
        )

    def test_add_top_level_fields(self):
        self.template.add_top_level_fields(
            ua.GoogleFieldType.TITLE_MODULE,
            image='https://s3.amazonaws.com/passtools_prod/1/images/default-loyalty-logo.png',
            imageDescription='An image!'
        )
        payload = self.template._create_payload()
        self.assertEqual(
            payload[ua.GoogleFieldType.TITLE_MODULE]['image'],
            'https://s3.amazonaws.com/passtools_prod/1/images/default-loyalty-logo.png'
        )
        self.assertEqual(
            payload[ua.GoogleFieldType.TITLE_MODULE]['imageDescription'],
            'An image!'
        )

    def test_set_logo_image(self):
        self.template.set_logo_image(
            'https://imgur.com/cool_image.png', description='Image'
        )
        payload = self.template._create_payload()
        self.assertEqual(
            payload[ua.GoogleFieldType.TITLE_MODULE]['image'],
            'https://imgur.com/cool_image.png'
        )

    def test_set_background_image(self):
        self.template.set_background_image(
            'https://imgur.com/cool_image.png',
            description='Super cool image'
        )
        payload = self.template._create_payload()
        self.assertEqual(
            payload[ua.GoogleFieldType._IMAGE_MODULE]['image'],
            'https://imgur.com/cool_image.png'
        )

    def test_set_offer(self):
        date = datetime.datetime(2017,1,31)
        offer_json = {
            'multiUserOffer': False,
            'redemptionChannel': 'both',
            'provider': 'The provider',
            'endTime': '2017-01-31T00:00'
        }
        self.template.set_offer(
            multi_user_offer=False,
            redemption_channel='both',
            provider='The provider',
            endtime=date
        )

        self.assertEqual(
            self.template.top_level_fields[ua.GoogleFieldType._OFFER_MODULE],
            offer_json
        )

    def test_add_message(self):
        # Test error
        self.assertRaises(
            ValueError,
            lambda: self.template.add_message(header="There's no body here!")
        )

        # Test minimal
        self.template.add_message(body='Test body')
        payload = self.template._create_payload()
        self.assertEqual(payload['messages'][0]['body'], 'Test body')

        # Test additional non-minimal message
        self.template.add_message(
            header='A title',
            body='Test body',
            action_uri='https://urbanairship.com'
        )
        payload = self.template._create_payload()
        self.assertEqual(len(payload['messages']), 2)
        self.assertEqual(payload['messages'][1]['header'], 'A title')

    def test_process_module_data(self):
        temp = ua.GoogleTemplate()
        module_data = {
            'Program Points': {
                'balance': 250,
                'label': 'Program Points',
                'hideEmpty': False,
                'formatType': 'Number',
                'fieldType': 'loyaltyPoints',
                'order': 1
            },
            'Tier': {
                'balance':  '2.0',
                'label': 'Tier',
                'hideEmpty': False,
                'formatType': 'Number',
                'fieldType': 'pointsModule',
                'order': 2
            }
        }

        program_points = ua.Field(
            name='Program Points',
            label='Program Points',
            value=250,
            hideEmpty=False,
            formatType='Number',
            fieldType=ua.GoogleFieldType.POINTS_MODULE,
            order=1
        )

        tier = ua.Field(
            name='Tier',
            label='Tier',
            value='2.0',
            hideEmpty=False,
            formatType='Number',
            fieldType=ua.GoogleFieldType.POINTS_MODULE,
            order=2
        )

        temp._process_module_data(ua.GoogleFieldType.POINTS_MODULE, module_data)
        self.assertTrue(temp.fields['Program Points'] == program_points)
        self.assertTrue(temp.fields['Tier'] == tier)

    def test_handle_reserved_names(self):
        temp = ua.GoogleTemplate()
        image_special = {
            'title.string': 'https://www.google.com',
            'hideEmpty': False,
            'formatType': 'String',
            'fieldType': 'titleModule',
            'description.string': ''
        }

        temp._handle_reserved_names(
            ua.GoogleFieldType.TITLE_MODULE,
            'image',
            image_special
        )

        self.assertEqual(
            temp.top_level_fields[ua.GoogleFieldType.TITLE_MODULE]['image'],
            'https://www.google.com'
        )



import json
import logging
from collections import defaultdict

import common
import uareach.fields as wf
from uareach import util


logger = logging.getLogger(__name__)


class TemplateMetadata(util.Constant):
    KEY_CLASS = 'TemplateMetadata'

    VENDOR = 'vendor'
    PROJECT_TYPE = 'project_type'
    TEMPLATE_TYPE = 'template_type'
    VENDOR_ID = 'vendorId'
    DELETED = 'deleted'
    DESCRIPTION = 'description'
    NAME = 'name'
    DISABLED = 'disabled'
    TEMPLATE_ID = 'template_id'
    PROJECT_ID = 'project_id'
    TYPE = 'type_'


class ProjectType(util.Constant):
    """Class representing possible projectTypes."""
    BOARDING_PASS = 'boardingPass'
    COUPON = 'coupon'
    EVENT_TICKET = 'eventTicket'
    GENERIC = 'generic'
    LOYALTY = 'loyalty'
    GIFT_CARD = 'giftCard'
    MEMBER_CARD = 'memberCard'


class TemplateType(util.Constant):
    """Class representing possible template types."""
    BOARDING_PASS = 'Boarding Pass'
    COUPON = 'Coupon'
    EVENT_TICKET = 'Event Ticket'
    GENERIC = 'Generic'
    STORE_CARD = 'Store Card'
    LOYALTY = 'Loyalty'
    OFFER = 'Offer'
    GIFT_CARD = 'Gift Card'


class Type(util.Constant):
    """These are template type, project type combinations that
    are known to work.

    .. note::

        There are currently discrepencies between Apple's and
        Google's acceptable (template)type and projectType
        pairs. These tuples represent Apple's pairing. For google,
        the _create_payload function under GoogleTemplate will perform
        the necessary conversions, ensuring the projectType and type
        keys are valid. In other words, these can be used for either
        Apple or Google templates.
    """
    LOYALTY = (TemplateType.STORE_CARD, ProjectType.LOYALTY)
    COUPON = (TemplateType.COUPON, ProjectType.COUPON)
    GIFT_CARD = (TemplateType.STORE_CARD, ProjectType.GIFT_CARD)
    MEMBER_CARD = (TemplateType.GENERIC, ProjectType.MEMBER_CARD)
    EVENT_TICKET = (TemplateType.EVENT_TICKET, ProjectType.EVENT_TICKET)
    BOARDING_PASS = (TemplateType.BOARDING_PASS, ProjectType.BOARDING_PASS)
    GENERIC = (TemplateType.GENERIC, ProjectType.GENERIC)


class TemplateHeader(util.Constant):
    KEY_CLASS = 'TemplateHeader'

    BACKGROUND_COLOR = 'background_color'
    BACKGROUND_IMAGE = 'background_image'
    BARCODE_ALT_TEXT = 'barcodeAltText'
    BARCODE_ENCODING = 'barcode_encoding'
    BARCODE_TYPE = 'barcode_type'
    BARCODE_VALUE = 'barcode_value'
    EXPIRATION_DATE = 'expirationDate'
    FOOTER_IMAGE = 'footer_image'
    FOREGROUND_COLOR = 'foreground_color'
    ICON_IMAGE = 'icon_image'
    LOGO_COLOR = 'logo_color'
    LOGO_IMAGE = 'logo_image'
    LOGO_TEXT = 'logo_text'
    STRIP_IMAGE = 'strip_image'
    SUPPRESS_STRIP_SHINE = 'suppress_strip_shine'
    THUMBNAIL_IMAGE = 'thumbnail_image'
    TRANSIT_TYPE = 'transitType'

    @classmethod
    def google_headers(cls):
        return [
            cls.BARCODE_ALT_TEXT,
            cls.BARCODE_TYPE,
            cls.BARCODE_ENCODING,
            cls.BARCODE_VALUE,
            cls.STRIP_IMAGE,
            cls.SUPPRESS_STRIP_SHINE,
            cls.EXPIRATION_DATE,
        ]


class BarcodeType(util.Constant):
    PDF_417 = 'PDF_417'
    AZTEC = 'AZTEC'
    QR_CODE = 'QR_CODE'
    CODE_128 = 'CODE_128'
    UPC_A = 'UPC_A'
    EAN_13 = 'EAN_13'
    CODE_39 = 'CODE_39'

    @classmethod
    def convert_to_apple(cls, value):
        if value == cls.PDF_417:
            return 'PKBarcodeFormatPDF417'
        elif value == cls.AZTEC:
            return 'PKBarcodeFormatAztec'
        elif value == cls.QR_CODE:
            return 'PKBarcodeFormatQR'
        elif value == cls.CODE_128:
            return 'PKBarcodeFormatCode128'
        else:
            raise ValueError("Invalid Apple barcode_type '{}'".format(value))


class TransitType(util.Constant):
    GENERIC = 'transitTypeGeneric'
    BUS = 'transitTypeBus'
    AIR = 'transitTypeAir'
    BOAT = 'transitTypeBoat'
    TRAIN = 'transitTypeTrain'


def get_template(reach, template_id=None, external_id=None):
    """Retrieve a template.

    Arguments:
        reach (reach.core.Reach): A uareach client object.
        template_id (str or int): The template ID of the template you wish
            to delete.
        external_id (str or int): The external ID of the template you wish
            to delete.

    Returns:
        An AppleTemplate or GoogleTemplate object.

    Raises:
        ValueError: If neither, or both, of template_id and external_id
            is specified.

    Example:
        >>> my_template = get_template(external_id=12345)
        <Template name:SummerSale, vendor:Apple>
    """
    if not (template_id or external_id) or (template_id and external_id):
        raise ValueError(
            'Please specify only one of template_id or external_id.'
        )

    response = reach.request(
        method='GET',
        body=None,
        url=Template.build_url(
            common.TEMPLATE_BASE_URL,
            main_id=template_id,
            external_id=external_id
        ),
        version=1.2
    )
    payload = response.json()
    logger.info('Successfully retrieved template: {}.'.format(
        template_id if template_id else external_id
    ))
    return _template_create_dispatch(payload)


def _template_create_dispatch(data):
    """Helper function, called by get_template. Determines whether
    the retrieved template is Apple or Google.

    Arguments:
        data (dict): A dictionary to be loaded into an AppleTemplate or
            GoogleTemplate object.
    Raises:
        ValueError: If the template vendor is not 'Apple' or 'Google'.
        KeyError: If the json template structure is unrecognized.
    Returns:
        An Apple or Google template.
    Example:
        >>> template_create_dispatch(template_dictionary)
        <Template name:SummerSale, vendor:Apple>
    """
    try:
        vendor = data['templateHeader']['vendor']
        if vendor == 'Apple':
            return AppleTemplate.from_data(data)
        elif vendor == 'Google':
            return GoogleTemplate.from_data(data)
        else:
            raise ValueError('Unrecognized vendor: {}'.format(vendor))
    except KeyError as k:
        logger.exception('Unrecognized template structure: {}'.format(data))
        raise


def delete_template(reach, template_id=None, external_id=None):
    """Delete a template.

    Arguments:
        reach (reach.core.Reach): A UA Reach object.
        template_id (str or int): The ID of the template you wish to delete.
        external_id (str or int): The external ID of the template you wish
            to delete.

    Returns:
        A boolean representing success.

    Raises:
        ValueError: If neither, or both, of template_id and external_id
            are specified.

    Example:
        >>> delete_template(ua_reach, template_id=123456)
        True
    """
    if not (template_id or external_id) or (template_id and external_id):
        raise ValueError(
            'Please specify only one of template_id or external_id.'
        )

    response = reach.request(
        method='DELETE',
        body=None,
        url=Template.build_url(
            common.TEMPLATE_BASE_URL,
            main_id=template_id,
            external_id=external_id
        ),
        version=1.2
    )
    logger.info('Successful template deletion: {}'.format(
        template_id if template_id else external_id
    ))
    return True


def duplicate_template(reach, template_id=None, external_id=None):
    """Duplicate a template.

    Arguments:
        reach (reach.core.Reach): A UA Reach object.
        template_id (str or int): The ID of the template you wish to delete.
        external_id (str or int): The external ID of the template you wish
            to delete.

    Returns:
        A dictionary containing the id of the newly created template.

    Raises:
        ValueError: If neither, or both, of template_id and external_id
            are specified.

    Example:
        >>> duplicate_template(ua_reach, template_id=12345)
        {'templateId': 12346}
    """
    if not (template_id or external_id) or (template_id and external_id):
        raise ValueError(
            'Please specify only one of template_id or external_id.'
        )

    response = reach.request(
        method='POST',
        body=None,
        url=Template.build_url(
            common.TEMPLATE_DUPLICATE_URL,
            main_id=template_id,
            external_id=external_id
        ),
        version=1.2
    )
    logger.info('Successfully created template copy: {}'.format(
        template_id if template_id else external_id
    ))
    return response.json()


def add_template_locations(
    reach, locations, template_id=None, external_id=None
):
    """Add locations to a template.

    Arguments:
        reach (reach.core.Reach): A uareach client object.
        locations (list of dicts): A list of location objects, represented
            as dictionaries.
        template_id (str or int): The ID of the template you wish to add
            locations to.
        external_id (str or int): The external ID of the template you wish to add
            locations to.

    Returns:
        A list of dicts, each containing the location value, locationId,
            and fieldId.

    Raises:
        ValueError: If neither, or both of, template_id and external_id
            are specified.

    Example:
        >>> add_template_locations(
        ...     ua_reach, [location1, location2], template_id=12345
        ... )
        [{'value': { '...' }, 'locationId': 12345, 'fieldId': 54321},
         {'value': { '...' }, 'locationId': 56789, 'fieldId': 98765}]
    """
    if not (template_id or external_id) or (template_id and external_id):
        raise ValueError(
            'Please specify only one of template_id or external_id.'
        )

    response = reach.request(
        method='POST',
        body=json.dumps({'locations': locations}),
        url=Template.build_url(
            common.TEMPLATE_ADD_LOCATION_URL,
            main_id=template_id,
            external_id=external_id
        ),
        content_type='application/json',
        version=1.2
    )
    logger.info("Successfully added {} locations to template '{}'.".format(
        len(locations),
        template_id if template_id else external_id
    ))
    return response.json()


def delete_template_location(
        reach, location_id, template_id=None, external_id=None
):
    """Remove a location from a template

    Arguments:
        reach (reach.core.Reach): A uareach client object.
        location_id (str or int): The ID of the location you wish to remove.
        template_id (str or int): The ID of the template you wish to remove
            locations from.
        external_id (str or int): The external ID of the template you wish
            to remove locations from.

    Returns:
        A boolean representing success.

    Raises:
        ValueError: If both or neither of template_id and external_id are
            specified.

    Example:
        >>> delete_template_location(ua_reach, 12345, template_id=44444)
        True
    """
    if not (template_id or external_id) or (template_id and external_id):
        raise ValueError(
            'Please specify only one of template_id or external_id.'
        )

    response = reach.request(
        method='DELETE',
        body=None,
        url=Template.build_url(
            common.TEMPLATE_REMOVE_LOCATION_URL,
            main_id=template_id,
            external_id=external_id,
            location_id=location_id
        ),
        version=1.2
    )
    logger.info("Successfully removed location '{}' from template '{}'.".format(
        location_id,
        template_id if template_id else external_id
    ))

    return True


class TemplateList(common.IteratorParent):
    """Forms a template listing request.

    Arguments:
        reach (reach.core.Reach): The uareach client object.
        page_size (int): The size of each page of the template listing
            response. Defaults to 10.
        page (int): The page to start the listing response on. Defaults
            to 1.
        order (string): The attribute used to order the templates. Can be one
            of ``"id"``, ``"name"``, ``"createdAt"``, or ``"updatedAt"``.
            Defaults to ``"id"``.
        direction (string): The direction of the template list, can be one of
            ``"ASC"`` or ``"DESC"``. Defaults to ``"DESC"``.

    Example:
        >>> template_list = TemplateList()
        >>> for item in template_list:
        ...     print item['createdAt']
        ...
        2016-07-27T21:14:47Z
        2016-07-23T12:13:32Z
        2016-06-09T17:32:14Z

    """
    base_url = common.TEMPLATE_BASE_URL.format('headers')
    data_attribute = 'templateHeaders'

    def __init__(
            self, reach, page_size=None, page=None, order=None, direction=None
    ):
        if direction not in {None, 'ASC', 'DESC'}:
            raise ValueError(
                'Unrecognized direction "{}". '
                'Please use one of "ASC" or "DESC"'.format(direction)
            )

        params = {
            'pageSize': page_size,
            'page': page,
            'order': order,
            'direction': direction
        }
        params = {k: v for k, v in params.iteritems() if v is not None}
        super(TemplateList, self).__init__(reach, params)


class Template(object):
    """Superclass for representing a uareach Template class. Should not be used
    directly -- use AppleTemplate or GoogleTemplate to perform template
    operations.
    """
    HEADERS = TemplateHeader
    HEADER_MAP = {
        TemplateHeader.BACKGROUND_COLOR:     {'fieldType': 'topLevel', 'formatType': 'String'},
        TemplateHeader.BACKGROUND_IMAGE:     {'fieldType': 'image', 'formatType': 'String'},
        TemplateHeader.BARCODE_ALT_TEXT:     {'fieldType': 'barcode', 'formatType': 'String'},
        TemplateHeader.BARCODE_ENCODING:     {'fieldType': 'barcode', 'formatType': 'String'},
        TemplateHeader.BARCODE_TYPE:         {'fieldType': 'barcode', 'formatType': 'String'},
        TemplateHeader.BARCODE_VALUE:        {'fieldType': 'barcode', 'formatType': 'String'},
        TemplateHeader.EXPIRATION_DATE:      {'fieldType': 'topLevel', 'formatType': 'String'},
        TemplateHeader.FOOTER_IMAGE:         {'fieldType': 'image', 'formatType': 'String'},
        TemplateHeader.FOREGROUND_COLOR:     {'fieldType': 'topLevel', 'formatType': 'String'},
        TemplateHeader.ICON_IMAGE:           {'fieldType': 'image', 'formatType': 'String'},
        TemplateHeader.LOGO_COLOR:           {'fieldType': 'topLevel', 'formatType': 'String'},
        TemplateHeader.LOGO_IMAGE:           {'fieldType': 'image', 'formatType': 'String'},
        TemplateHeader.LOGO_TEXT:            {'fieldType': 'topLevel', 'formatType': 'String'},
        TemplateHeader.STRIP_IMAGE:          {'fieldType': 'image', 'formatType': 'String'},
        TemplateHeader.SUPPRESS_STRIP_SHINE: {'fieldType': 'topLevel', 'formatType': 'String'},
        TemplateHeader.THUMBNAIL_IMAGE:      {'fieldType': 'image', 'formatType': 'String'},
        TemplateHeader.TRANSIT_TYPE:         {'fieldType': 'passTop', 'formatType': 'String'},
    }

    METADATA = TemplateMetadata
    READ_ONLY_METADATA = ['createdAt', 'template_id', 'updatedAt', 'project_id']

    def __init__(self):
        self.headers = {}
        self.fields = defaultdict(dict)
        self.metadata = {}

    def create(self, reach, project_id=None, external_id=None):
        """Submits a template create request to the API.

        Arguments:
            reach (reach.core.Reach): A uareach client object.
            project_id (str or int): A project ID.
            external_id (str or int): An external ID.

        Returns:
            A dict containing the template ID.

        Example:
            >>> template.create_template(ua_reach, project_id=12345)
            {'templateId': 54321}
        """
        if not project_id:
            try:
                project_id = self.metadata[TemplateMetadata.PROJECT_ID]
            except KeyError:
                raise ValueError(
                    "Either set project_id when calling update, "
                    "or set the project_id attribute via the add_metadata "
                    "method, e.g., your_template.add_metadata(project_id=1234)."
                )

        payload = self._create_payload()
        self._validate(payload)
        response = reach.request(
            method='POST',
            body=json.dumps(payload),
            url=Template.build_url(
                common.TEMPLATE_BASE_URL,
                main_id=project_id,
                external_id=external_id
            ),
            content_type='application/json',
            version=1.2
        )
        template_id = response.json().get('templateId')
        logger.info('Successful template creation -- ID: {}, EXTERNAL_ID: {}'.format(
            template_id, external_id
        ))
        self.metadata[TemplateMetadata.TEMPLATE_ID] = template_id
        return {'templateId': template_id}

    def update(self, reach, template_id=None, external_id=None):
        """Submits a template update request to the API.

        Arguments:
            reach (reach.core.Reach): A uareach client object.
            template_id (str or int): A template ID.
            external_id (str or int): An external ID.

        Returns:
            A boolean representing success.

        Example:
            >>> template.update(ua_reach)
            True
        """
        if template_id and external_id:
            raise ValueError('Only set one of template_id or external_id.')
        if not external_id and not template_id:
            try:
                template_id = self.metadata[TemplateMetadata.TEMPLATE_ID]
            except KeyError:
                raise ValueError(
                    "Either set template_id or external_id when calling "
                    "update, or set the template id attribute via the "
                    "add_metadata method, e.g., "
                    "your_template.add_metadata(template_id=1234)."
                )

        payload = self._create_payload()
        self._validate(payload)
        response = reach.request(
            method='PUT',
            body=json.dumps(payload),
            url=Template.build_url(
                common.TEMPLATE_BASE_URL,
                main_id=template_id,
                external_id=external_id
            ),
            content_type='application/json',
            version=1.2
        )
        logger.info('Successful template creation: {}'.format(
            template_id if template_id else external_id
        ))
        return True

    @classmethod
    def from_data(cls, data):
        """Create a template/pass from JSON data.

        Arguments:
            data (dict): The JSON data to convert to a template or pass
                object.

        Returns:
            An Apple or Google template object.

        Example:
            >>> loyalty_template = AppleTemplate.from_data({
            ...     'name': 'Loyalty Template',
            ...     'description': 'A loyalty template',
            ...     'fieldsModel': {
            ...         'fields': {
            ...             ...
            ...         },
            ...         'headers': {
            ...             ...
            ...         }
            ...     },
            ...     ...,
            ...     'vendor': 'Apple'
            ... })
            <Template name:Loyalty Template, vendor:Apple>
        """
        template = cls()
        # Handle Metadata
        template.metadata = data.get('templateHeader', {})
        # Convert id values to descriptive keys
        if template.metadata.get('id'):
            template.metadata[TemplateMetadata.TEMPLATE_ID] = template.metadata['id']
            del template.metadata['id']
        if template.metadata.get('projectId'):
            template.metadata[TemplateMetadata.PROJECT_ID] = template.metadata['projectId']
            del template.metadata['projectId']

        # Handle headers
        template.headers = data.get('fieldsModel', {}).pop('headers', {})
        if template.headers.get(TemplateHeader.BARCODE_TYPE):
            no_pk = template.headers[TemplateHeader.BARCODE_TYPE]['value'].replace('PKB', 'B')
            template.headers[TemplateHeader.BARCODE_TYPE]['value'] = no_pk

        # Handle user_locations
        if data.get('userlocations'):
            template.user_locations = data.pop('userlocations')
        return template

    """ Field manipulation """
    def add_fields(self, *args):
        """Add uareach.fields.Field objects.

        Arguments:
            args (reach.fields.Field): Field objects to add to your template or
                pass.

        Example:
            >>> points_field = Field(name='Rewards Points', fieldType='primary')
            >>> member_field = Field(name='Member Name', fieldType='secondary')
            >>> template_or_pass.add_fields(points_field, member_field)
        """
        for field in args:
            self.fields[field.name] = field

    def remove_fields(self, *args):
        """Removes uareach.fields.Field objects.

        Arguments:
            args (string): The names of the fields to be deleted, e.g.,
                'Rewards Points'/'Member Name'/'Background'

        Example:
            >>> template_or_pass.remove_fields('Rewards Points', 'Member Name')
        """
        for name in args:
            del self.fields[name]

    def set_fields(self, *args):
        """Set the uareach.fields.Field objects. Will overwrite any existing
        fields set on the template or pass.

        Arguments:
            *args (reach.fields.Field): Field objects to add to your template or
                pass.

        Example:
            >>> points_field = Field(name='Rewards Points', fieldType='primary')
            >>> member_field = Field(name='Member Name', fieldType='secondary')
            >>> template_or_pass.set_fields(points_field, member_field)
        """
        self.fields.clear()
        self.add_fields(*args)

    """ Metadata manipulation. """
    def add_metadata(self, **kwargs):
        """Add metadata items.

        .. note::

            If working with a template, acceptable keys are accessible through
            TemplateMetadata. Likewise, if working with a pass, acceptable keys
            are accessible through PassMetadata.

        Arguments:
            **kwargs: Key, value pairs specifying the metadata items to add.

        Example:
            >>> template.add_metadata(
            ...     project_type='memberCard',
            ...     template_type='Store Card',
            ...     description='Hello this is a description'
            ... )
            >>> # method is identical for passes, but acceptable keys differ
            >>> pass_.add_metadata(
            ...     pass_id=12345
            ... )
        """
        for header, value in kwargs.iteritems():
            self.METADATA.validate(header)
            self._add_metadata_helper(header, value)

    def remove_metadata(self, *args):
        """Remove metadata items

        .. note::

            If working with a template, acceptable strings are accessible through
            TemplateMetadata. Likewise, if working with a pass, acceptable strings
            are accessible through PassMetadata.

        Arguments:
            *args (strings): Keys of the metadata values to remove.

        Example:
            >>> my_template.remove_metadata('projectType', 'description')
            >>> my_pass.remove_metadata('pass_id')
        """
        for name in args:
            del self.metadata[name]

    def set_metadata(self, **kwargs):
        """Remove all currently set metadata items and then add the
        specified metadata

        .. note::

            If working with a template, acceptable keys are accessible through
            TemplateMetadata. Likewise, if working with a pass, acceptable keys
            are accessible through PassMetadata.

        Arguments:
            **kwargs: Key, value pairs specifying the metadata items to add.

        Example:
            >>> template.set_metadata(
            ...     project_type='memberCard',
            ...     template_type='Store Card',
            ...     description='Hello this is a description'
            ... )
            >>> # method is identical for passes, but acceptable keys differ
            >>> pass_.set_metadata(
            ...     pass_id=12345
            ... )
        """
        self.metadata.clear()
        self.add_metadata(**kwargs)

    def _add_metadata_helper(self, name, value):
        """Does validation on metadata (key, val) pairs, then set metadata
        if it passes validation.
        """
        self.METADATA.validate(name)
        if name == self.METADATA.TEMPLATE_TYPE:
            self.metadata['type'] = value
        elif name == self.METADATA.PROJECT_TYPE:
            self.metadata['projectType'] = value
        elif name == self.METADATA.TYPE:
            Type.validate(value)
            self.metadata['type'] = value[0]
            self.metadata['projectType'] = value[1]
        else:
            self.metadata[name] = value

    """ Header manipulation. """
    def add_header(self, header, value, format_type=None, field_type=None):
        """Add a header.

        Arguments:
            header (str): The header, represented as a string
                (e.g. 'logo_color')
            value (str): The header value.

        Raises:
            ValueError: If the `header` is not contained in Header.values

        Example:
            >>> template_or_pass.add_header('barcode_value', '123456789')
        """
        self.HEADERS.validate(header)
        if header == self.HEADERS.BARCODE_TYPE:
            BarcodeType.validate(value)

        # Use the default formatType and fieldType values if not specified.
        if format_type is None:
            format_type = self.HEADER_MAP[header]['formatType']
        if field_type is None:
            field_type = self.HEADER_MAP[header]['fieldType']

        self.headers[header] = {
            'formatType': format_type,
            'fieldType': field_type,
            'value': value
        }

    def add_headers(self, **kwargs):
        """Add multiple headers.

        Arguments:
            **kwargs: Key, value pairs specifying header values.

        Example:
            >>> template_or_pass.add_headers(
            ...     barcode_value='iso-8859-1',
            ...     logo_image='https://google.com/fun.png'
            ... )
        """
        for header, value in kwargs.iteritems():
            self.add_header(header, value)

    def remove_headers(self, *args):
        """Remove headers.

        Arguments:
            *args (list of strings): A list of header values to remove.

        Example:
            >>> template_or_pass.remove_headers('logo_color', 'icon_image')
        """
        for header in args:
            del self.headers[header]

    def set_headers(self, **kwargs):
        """Remove all currently set headers and then add the
        specified headers.

        Arguments:
            **kwargs: Header specifications. Note that the keywords must be one
                of the keys listed in Header.values

        Example:
            >>> template.set_headers(
            ...     barcode_value='iso-8859-1',
            ...     logo_image='https://google.com/fun.png'
            ... )
        """
        self.headers.clear()
        self.add_headers(**kwargs)

    def view(self):
        """View the entirety of the template or pass, including keys that
        are read-only.

        Returns:
            A dict representing the template or pass.

        Example:
            >>> my_template.view()
            {'headers': {...}, 'fields': {...}, 'key1': 'val1', ... }
        """
        payload = {'headers': self.headers}
        payload.update(self.metadata)
        return payload

    def _create_payload(self):
        """Formats the payload for a request. Filters out read-only items,
        and handles formatting differences between Apple and Google.
        """
        payload = {'headers': {}}
        for key, val in self.headers.iteritems():
            payload['headers'][key] = val
        for key, val in self.metadata.iteritems():
            if key not in self.READ_ONLY_METADATA:
                payload[key] = val
        return payload

    @staticmethod
    def _validate(payload):
        pass

    @staticmethod
    def build_url(url, main_id=None, external_id=None, location_id=None):
        if location_id and main_id:
            return url.format(main_id, location_id)
        elif location_id:
            return url.format('id/' + str(external_id), location_id)
        elif main_id and external_id:
            return url.format(str(main_id) + '/id/' + str(external_id))
        elif main_id:
            return url.format(main_id)
        else:
            return url.format('id/' + str(external_id))


@util.inherit_docs
class AppleTemplate(Template):
    """Represents an Apple template object."""

    def __init__(self):
        super(AppleTemplate, self).__init__()
        self.beacons = []
        self.metadata['vendor'] = 'Apple'
        self.metadata['vendorId'] = 1

    @classmethod
    def from_data(cls, data):
        template = super(AppleTemplate, cls).from_data(data)
        fields = data.get('fieldsModel', {}).get('fields', {})
        for name, json_field in fields.iteritems():
            template.fields[name] = wf.Field.build_apple_field(
                name, json_field
            )
        template.beacons = data.get('beacons', [])
        return template

    def view(self):
        payload = super(AppleTemplate, self).view()
        payload.update({'fields': {}, 'beacons': []})
        for name, field in self.fields.iteritems():
            field._build_common_template_json()
            payload['fields'][name] = field.build_apple_json()
        payload['beacons'] = self.beacons
        return payload

    def add_header(self, header, value, format_type=None, field_type=None):
        # Check Apple-specific enums
        if header == self.HEADERS.TRANSIT_TYPE:
            TransitType.validate(value)

        super(AppleTemplate, self).add_header(
            header, value, format_type, field_type
        )

        # Apple-specific validation
        if header == TemplateHeader.BARCODE_TYPE:
            try:
                self.headers[header]['value'] = BarcodeType.convert_to_apple(value)
            except ValueError:
                del self.headers[header]
                raise

    def _create_payload(self):
        payload = self.view()
        return {key: val for key, val in payload.iteritems() if
                key not in self.READ_ONLY_METADATA}

    def set_logo_image(self, value):
        """Set the logo image.

        Notes:
            This is mainly here to provide parity with the GoogleTemplate
            class. You could equivalently do template_or_pass.set_headers(logo_image='image')

        Arguments:
            value (str): A URL path to an image.

        Example:
            >>> template_or_pass.set_logo_image('https://urbanairship.com/cool_image.png')
        """
        self.set_headers(logo_image=value)

    # Apple-specific methods
    def add_beacon(self, uuid, relevant_text=None, major=None, minor=None):
        """Add a beacon to the template.

        Arguments:
            uuid (str): An identifier for this location.
            relevant_text (str): Text to display when a user approaches the
                location.
            major (int): A major location identifier.
            minor (int): A minor location identifier.

        Example:
            >>> template.add_beacon(
            ...     uuid='3526dee6-4ea8-11e6-beb8-9e71128cae77',
            ...     relevant_text='You are near something cool.',
            ...     major=2,
            ...     minor=34
            ... )
        """
        payload = {
            'uuid': uuid,
            'relevantText': relevant_text,
            'major': major,
            'minor': minor
        }

        self.beacons.append({
            key: val for key, val in payload.iteritems() if val is not None
        })

    def remove_beacon(self, uuid):
        """Remove a beacon from a template.

        Arguments:
            uuid (str): The UUID of the beacon you wish to remove.

        Raises:
            ValueError: When the uuid does not match any beacon associated
                with the template

        Example:
            >>> template.remove_beacon('8054f07e-238f-439e-93eb-0c2fe6829541')
        """
        for index, beacon in enumerate(self.beacons):
            if beacon['uuid'] == uuid:
                del self.beacons[index]
                return
        raise ValueError("Beacon with UUID {} not found".format(uuid))


@util.inherit_docs
class GoogleTemplate(Template):
    """Represents a Google template object."""

    # Some modules have reserved names (names that have special meaning and
    # cannot be used for field names).
    RESERVED_NAMES = {
        wf.GoogleFieldType.INFO_MODULE: ['hexFontColor', 'hexBackgroundColor'],
        wf.GoogleFieldType.TITLE_MODULE: ['image'],
        wf.GoogleFieldType._IMAGE_MODULE: ['imageModulesData']
    }

    def __init__(self):
        super(GoogleTemplate, self).__init__()
        self.top_level_fields = defaultdict(dict)
        self.messages = []
        self.metadata['vendor'] = 'Google'
        self.metadata['vendorId'] = 2

    # Override methods
    @classmethod
    def from_data(cls, data):
        template = super(GoogleTemplate, cls).from_data(data)
        modules = data.get('fieldsModel', {})
        for field_type, module_data in modules.iteritems():
            if field_type == 'vendor':
                continue
            # Handling special modules
            elif field_type == wf.GoogleFieldType._OFFER_MODULE:
                template.add_top_level_fields(
                    wf.GoogleFieldType._OFFER_MODULE,
                    **module_data
                )
            elif field_type == wf.GoogleFieldType._MESSAGE_MODULE:
                template.messages = module_data
            elif field_type == wf.GoogleFieldType._IMAGE_MODULE:
                if module_data.get(wf.GoogleFieldType._IMAGE_MODULE):
                    inner_module = module_data[wf.GoogleFieldType._IMAGE_MODULE]
                    template.add_top_level_fields(
                        wf.GoogleFieldType._IMAGE_MODULE,
                        **inner_module
                    )
            else:
                template._process_module_data(field_type, module_data)
        return template

    def view(self):
        payload = super(GoogleTemplate, self).view()
        payload.update({'fields': {}, 'messages': []})
        for name, field in self.fields.iteritems():
            field._build_common_template_json()
            payload['fields'][name] = field.build_google_template_json()
        payload['top_level_fields'] = self.top_level_fields
        payload['messages'] = self.messages
        return payload

    def _create_payload(self):
        payload = super(GoogleTemplate, self)._create_payload()

        # Handle type conversions
        if payload['projectType'] == ProjectType.COUPON:
            payload['type'] = TemplateType.OFFER
        if payload['projectType'] == ProjectType.GIFT_CARD:
            payload['type'] = TemplateType.GIFT_CARD
        if payload['projectType'] == ProjectType.LOYALTY:
            payload['type'] = TemplateType.LOYALTY

        # Handle standard fields
        for name, field in self.fields.iteritems():
            if not payload.get(field['fieldType']):
                payload[field['fieldType']] = {}
            field._build_common_template_json()
            payload[field['fieldType']][field.name] = field.build_google_template_json()

        # Handle top-level fields
        for field_type, fields in self.top_level_fields.iteritems():
            if not payload.get(field_type):
                payload[field_type] = {}
            payload[field_type].update(fields)

        # Handle messages
        if self.messages:
            payload['messages'] = self.messages

        return payload

    def add_header(self, header, value, format_type=None, field_type=None):
        super(GoogleTemplate, self).add_header(
            header, value, format_type, field_type
        )

        if header not in self.HEADERS.google_headers():
            del self.headers[header]
            raise ValueError(
                "The header '{}' is not used with "
                "Google templates".format(header)
            )

    def set_logo_image(self, value, description=None):
        """Set the logo image.

        Arguments:
            value (string): A URL pointing to an image.
            description (string): An optional string describing the logo
                image.

        Example:
            >>> template.set_logo_image(
            ...     'https://imgur.com/cool_image.png',
            ...     description='A super cool image'
            ... )
        """
        if not description:
            description = ''

        self.add_top_level_fields(
            wf.GoogleFieldType.TITLE_MODULE,
            image=value,
            imageDescription=description
        )

    def set_background_image(self, value, description=None):
        """Set the background image.

        Arguments:
            value (string): A URL pointing to an image.
            description (string): An optional string describing the background
                image.

        Example:
            >>> template.set_background_image(
            ...     'https://imgur.com/cool_image.png',
            ...     description='A super cool image'
            ... )
        """
        if not description:
            description = ''

        self.add_top_level_fields(
            wf.GoogleFieldType._IMAGE_MODULE,
            **{'image': value, 'imageDescription': description}
        )

    # Google-specific methods
    def add_top_level_fields(self, field_type, **kwargs):
        """Add top level values to the specified module location.

        Arguments:
            location (str): The module being added to.
            **kwargs: A mapping of string keys to string values.

        Raises:
            ValueError: If the location is not specified

        Example:
            >>> template.add_top_level_fields(
            ...     GoogleFieldType.TITLE_MODULE,
            ...     image="https://google.com/images/asdfoi1.png",
            ...     imageDescription="Logo Image"
            ... )

        """
        wf.GoogleFieldType.validate(field_type, allow_private=True)
        for key, val in kwargs.iteritems():
            self.top_level_fields[field_type][key] = val

    def set_offer(
        self,
        multi_user_offer=None,
        endtime=None,
        provider=None,
        redemption_channel=None
    ):
        """Set the offerModule.

        Arguments:
            * multi_user_offer (bool): One of True or False. Indicates
              whether the offer can be used by multiple users.
            * endtime (datetime.datetime): The offer expiration date.
            * provider (string): The provider name of the offer.
            * redemption_channel (string): Can be one of ``'online'``,
              ``'instore'``, ``'both'``, or ``'temporaryPriceReduction'``.

        Example:
            >>> template.set_offer(
            ...     multi_user_offer=False,
            ...     provider='UA'
            ...     redemption_channel='both',
            ... )
        """
        valid_chans = {'instore', 'online', 'both', 'temporaryPriceReduction'}
        if redemption_channel and redemption_channel not in valid_chans:
            raise ValueError(
                "Unrecognized redemption_channel '{}'. Valid \
                redemption_channels are: {}.".format(
                    redemption_channel, ', '.join(valid_chans)
                )
            )

        payload = {
            'multiUserOffer': multi_user_offer,
            'endTime': endtime.strftime('%Y-%m-%dT%H:%M') if endtime else None,
            'provider': provider,
            'redemptionChannel': redemption_channel
        }
        self.add_top_level_fields(
            wf.GoogleFieldType._OFFER_MODULE,
            **{key: val for key, val in payload.iteritems() if val != None}
        )

    def add_message(
        self,
        body=None,
        header=None,
        action_uri=None,
        action_uri_description=None,
        image_uri=None,
        image_description=None,
        starttime=None,
        endtime=None
    ):
        """Add a message object.

        Arguments:
            * body (string): The message body.
            * header (string): The message header.
            * action_uri (string): The URI to which users are directed
              upon clicking the message.
            * action_uri_description (string): Description for the
              ``action_uri``.
            * image_uri (string): Specify an image to display with the
              message.
            * image_description (string): Description for the image.
            * starttime (datetime.datetime): Valid ISO8805 date for start
              time of a message.
            * endtime (datetime.datetime): Valid ISO8805 date for end time of
              a message.

        Example:
            >>> template.add_message(
            ...     body='A message body',
            ...     header='A message header'
            ... )
        """
        if not body:
            raise ValueError("The 'body' key must be specified for each message.")

        payload = {
            'body': body,
            'header': header,
            'actionUri': action_uri,
            'actionUriDescription': action_uri_description,
            'imageUri': image_uri,
            'imageDescription': image_description,
            'startTime': starttime.strftime('%Y-%m-%dT%H:%M') if starttime else None,
            'endTime': endtime.strftime('%Y-%m-%dT%H:%M') if endtime else None
        }
        self.messages.append(
            {key: val for key, val in payload.iteritems() if val != None}
        )

    def _process_module_data(self, field_type, module_data):
        """Handle each module (e.g. textModulesData, etc).
        """
        for name, data in module_data.iteritems():
            if name in GoogleTemplate.RESERVED_NAMES.get(field_type, []):
                self._handle_reserved_names(field_type, name, data)
            elif isinstance(data, dict):
                self.fields[name] = wf.Field.build_google_field(name, data)
            elif isinstance(data, str):
                self.add_top_level_field(field_type, name, data)
            else:
                ValueError('Unrecognized Android template structure.')

    def _handle_reserved_names(self, field_type, name, data):
        """Handles specialized module data.
        """
        value = data.get(wf.Field.GOOGLE_VALUE_MAP[field_type])
        if name == 'hexFontColor':
            self.add_top_level_field(field_type, hexFontColor=value)
        elif name == 'hexBackgroundColor':
            self.add_top_level_field(field_type, hexBackgroundColor=value)
        elif name == 'image':
            self.add_top_level_fields(field_type, image=data['title.string'])
            if 'description.string' in data:
                self.add_top_level_fields(
                    field_type, imageDescription=data['description.string']
                )
        else:
            raise ValueError('Unrecognized reserved name: {}'.format(name))

    @staticmethod
    def _validate_payload(payload):
        """Validates the structure of a Google Template.
        """
        if not payload.get(ua.GoogleFieldType.TITLE_MODULE):
            raise ValueError('Must add a title field to the template.')
        if not payload.get(ua.GoogleFieldType.TITLE_MODULE, {}).get('image'):
            raise ValueError("Must use set_logo_image to set your template's logo image.")



import datetime
import json
import unittest

import mock
import requests

import pass_builders
import uareach as ua
from uareach.passes import Pass


class PassApiTest(unittest.TestCase):

    def setUp(self):
        self.client = ua.Reach('fake', 'creds')
        super(PassApiTest, self).setUp()

    @mock.patch.object(ua.Reach, '_request')
    def test_delete(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        mock_request.return_value = response

        reach_response = ua.delete_pass(self.client, pass_id=12345)
        mock_request.assert_called_with(
            'DELETE',
            None,
            ua.common.PASS_BASE_URL.format(12345),
            None,
            1.2,
            None
        )
        self.assertEqual(reach_response, True)

    @mock.patch.object(ua.Reach, '_request')
    def test_add_pass_locations(self, mock_request):
        locations = [
            {
                "longitude": -122.374,
                "latitude": 37.618,
                "relevantText": "Hello loc0",
                "streetAddress1": "address line #1",
                "streetAddress2": "address line #2",
                "city": "Palo Alto",
                "region": "CA",
                "regionCode": "94404",
                "country": "US"
            },
            {
                "longitude": -122.374,
                "latitude": 37.618,
                "relevantText": "Hello loc0",
                "streetAddress1": "address line #1",
                "streetAddress2": "address line #2",
                "city": "Palo Alto",
                "region": "CA",
                "regionCode": "94405",
                "country": "US"
            }
        ]
        body = json.dumps({
            'locations': locations
        })
        response = requests.Response()
        response._content = json.dumps([
            {
                'value': locations[0],
                'passLocationId': 231
            },
            {
                'value': locations[1],
                'passLocationId': 312
            }
        ])
        mock_request.return_value = response

        reach_response = ua.add_pass_locations(
            self.client, locations, pass_id=12345
        )
        mock_request.assert_called_with(
            'POST',
            body,
            ua.common.PASS_ADD_LOCATION_URL.format(12345),
            'application/json',
            1.2,
            None
        )
        self.assertEqual(reach_response[0]['value'], locations[0])
        self.assertEqual(reach_response[0]['passLocationId'], 231)
        self.assertEqual(reach_response[1]['value'], locations[1])
        self.assertEqual(reach_response[1]['passLocationId'], 312)

    @mock.patch.object(ua.Reach, '_request')
    def test_delete_pass_location(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        mock_request.return_value = response

        reach_response = ua.delete_pass_location(
            self.client, 'location123', pass_id=12345
        )
        mock_request.assert_called_with(
            'DELETE',
            None,
            ua.common.PASS_DELETE_LOCATION_URL.format(
                12345, 'location123'
            ),
            None,
            1.2,
            None
        )

        self.assertEqual(reach_response, True)

    @mock.patch.object(ua.Reach, '_request')
    def test_get_pass(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        pass_json = json.dumps({
            'fields': {
                'Points': {
                    'fieldType': 'primary',
                    "numberStyle": 'numberStyleDecimal',
                    'value': 33.0
                }
            },
            'headers': {
                'barcode_encoding': {
                    'fieldType': 'barcode',
                    'formatType': 1,
                    'value': 'iso-8859-1'
                },
            },
            'id': '18971172',
            'templateId': '51234',
            'status': 'not_been_installed',
            'tags': []
        })
        response._content = pass_json
        mock_request.return_value = response
        pass_ = ua.get_pass(self.client, pass_id=12345)
        mock_request.assert_called_with(
            'GET',
            None,
            ua.common.PASS_BASE_URL.format(12345),
            None,
            1.2,
            None
        )

    def test_pass_dispatch(self):
        min_apple_json = {
            'fields': {
                'Points': {
                    'fieldType': 'primary',
                    "numberStyle": 'numberStyleDecimal',
                    'value': 33.0
                }
            }
        }
        apple_pass = ua.passes._pass_dispatch(min_apple_json)
        self.assertEquals(apple_pass.__class__.__name__, 'ApplePass')

        min_google_json = {
            'fields': {
                'Points': {
                    'fieldType': 'pointsModule',
                    'value': 33.0
                }
            }
        }
        google_pass = ua.passes._pass_dispatch(min_google_json)
        self.assertEquals(google_pass.__class__.__name__, 'GooglePass')

    @mock.patch.object(ua.Reach, '_request')
    def test_list_passes(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        pass_json = json.dumps({
            "count": 133,
            "passes":[
                {
                   "id":"12345",
                   "templateId":"12345",
                   "updatedAt":"2013-06-27T20:58:20.000Z",
                   "createdAt":"2013-06-27T20:58:18.000Z",
                   "serialNumber":"ff9db2ed-37aa-4691-a3b6-240108ac31f9",
                   "url":"https://wallet-api.urbanairship.com/v1/pass/1273/download"
                },
                {
                   "id":"12346",
                   "templateId":"12346",
                   "updatedAt":"2013-06-27T20:56:07.000Z",
                   "createdAt":"2013-06-27T20:56:03.000Z",
                   "serialNumber":"a2ea965a-0917-461d-ac67-749639831818",
                   "url":"https://wallet-api.urbanairship.com/v1/pass/1272/download"
                }
            ],
            "pagination":{
                "order": "id",
                "page": 1,
                "start": 0,
                "direction": "DESC",
                "pageSize": 2
            }
        })
        response._content = pass_json
        mock_request.return_value = response
        listing_response = ua.PassList(self.client)
        self.assertEqual(
            listing_response.next()['id'],
            '12345'
        )
        mock_request.assert_called_with(
            'GET',
            None,
            ua.common.PASS_BASE_URL.format(''),
            None,
            1.2,
            {}
        )


class PassTest(unittest.TestCase):

    def setUp(self):
        pass_ = Pass()
        member_name = ua.Field(
            name='Member Name',
            value='First Last'
        )
        pass_.add_fields(member_name)
        self.pass_ = pass_
        self.client = ua.Reach('fake', 'creds')

    @mock.patch.object(ua.Reach, '_request')
    def test_create_pass(self, mock_request):
        response = requests.Response()
        response._content = json.dumps({
            'url': 'https://wallet-api.urbanairship.com/v1/pass/123456/download',
            'id': '12345678'
        })
        response.status_code = 200
        mock_request.return_value = response
        payload = self.pass_._create_payload()
        create_response = self.pass_.create(self.client, template_id=54321)
        mock_request.assert_called_with(
            'POST',
            json.dumps(payload),
            ua.common.PASS_BASE_URL.format(54321),
            'application/json',
            1.2,
            None
        )
        self.assertEqual(self.pass_.metadata['id'], '12345678')

    @mock.patch.object(ua.Reach, '_request')
    def test_update_pass(self, mock_request):
        response = requests.Response()
        response.status_code = 200
        response._content = json.dumps({'ticketId': 12345678})
        mock_request.return_value = response
        update_response = self.pass_.update(self.client, pass_id=12345)

        mock_request.assert_called_with(
            'PUT',
            json.dumps(self.pass_._create_payload()),
            ua.common.PASS_BASE_URL.format(12345),
            'application/json',
            1.2,
            None
        )
        self.assertEqual(update_response['ticketId'], 12345678)

    def test_create_payload(self):
        pass_ = Pass()
        member_name = ua.Field(
            name='Member Name',
            value='Max Del Giudice'
        )
        member_id = ua.Field(
            name='Member ID',
            value=12345678
        )
        points = ua.Field(
            name='Points',
            value=5555
        )
        pass_.add_fields(member_name, member_id, points)
        pass_.add_headers(barcode_value='99999999')
        self.assertEqual(
            pass_._create_payload(),
            {
                'fields': {
                    'Member Name': {'value': 'Max Del Giudice'},
                    'Member ID': {'value': 12345678},
                    'Points': {'value': 5555}
                },
                'headers': {
                    'barcode_value': {
                        'fieldType': 'barcode',
                        'formatType': 'String',
                        'value': '99999999'
                    }
                }
            }
        )

    def test_set_public_url(self):
        self.pass_.set_public_url(type_='single')
        self.assertEqual(
            self.pass_._create_payload()['publicUrl'],
            {'type': 'single'}
        )
        self.pass_.set_public_url(type_='multiple')
        self.assertEqual(
            self.pass_._create_payload()['publicUrl'],
            {'type': 'multiple'}
        )
        self.assertRaises(
            ValueError, self.pass_.set_public_url, type_='none'
        )


class ApplePassTest(unittest.TestCase):

    def setUp(self):
        self.pass_ = pass_builders.build_apple_pass()
        self.client = ua.Reach('fake', 'creds')

    def test_add_remove_beacons(self):
        beacon = {
            'uuid': 'abcdef',
            'relevantText': "Here's a beacon",
            'major': 132,
            'minor': 32
        }

        self.pass_.add_beacon(
            'abcdef',
            relevant_text="Here's a beacon",
            major=132,
            minor=32
        )
        self.assertEqual(
            self.pass_._apple_features.beacons,
            [beacon]
        )

    def test_set_expiration(self):
        self.pass_.set_expiration(datetime.datetime(2014, 8, 6))
        json_payload = self.pass_._create_payload()
        self.assertEqual(len(json_payload['headers']), 2)
        self.assertEqual(
            json_payload['headers']['expirationDate'],
            {'value': '2014-08-06T00:00'}
        )

    def test_set_logo_image(self):
        self.pass_.set_logo_image('https://urbanairship.com/cool_image.png')
        self.assertEqual(
            self.pass_._create_payload()['headers']['logo_image'],
            {
                'value': 'https://urbanairship.com/cool_image.png',
                'fieldType': 'image',
                'formatType': 'String'
            }
        )

class GooglePassTest(unittest.TestCase):

    def setUp(self):
        self.pass_ = pass_builders.build_google_pass()
        self.client = ua.Reach('fake', 'creds')

    def test_set_expiration(self):
        self.pass_.set_expiration(datetime.datetime(2014, 8, 6))
        json_payload = self.pass_._create_payload()

        self.assertEqual(
            json_payload['fields']['endTime'],
            {
                'value': '2014-08-06T00:00'
            }
        )

    def test_field_json(self):
        view_json = {
            "Program Points": {
                "label": "UPDATED_LABEL",
                "value": "UPDATED_VALUE"
            },
            "Program Details": {
                "label": "UPDATED_LABEL",
                "value": "UPDATED_VALUE"
            },
            "Tier": {
                "label": "UPDATED_LABEL",
                "value": "UPDATED_VALUE"
            },
            "Merchant Website": {
                "label": "UPDATED_LABEL",
                "value": "UPDATED_VALUE"
            },
            "Loyalty Program Name": {
                "label": "UPDATED_LABEL",
                "value": "UPDATED_VALUE"
            }
        }

        program_points = ua.Field(name='Program Points', label='UPDATED_LABEL', value='UPDATED_VALUE')
        program_details = ua.Field(name='Program Details', label='UPDATED_LABEL', value='UPDATED_VALUE')
        tier = ua.Field(name='Tier', label='UPDATED_LABEL', value='UPDATED_VALUE')
        merchant_website = ua.Field(name='Merchant Website', label='UPDATED_LABEL', value='UPDATED_VALUE')
        program_name = ua.Field(name='Loyalty Program Name', label='UPDATED_LABEL', value='UPDATED_VALUE')
        self.pass_.set_fields(
            program_points, program_details, tier, merchant_website, program_name
        )

        self.assertEquals(self.pass_._create_payload()['fields'], view_json)
        self.assertEquals(self.pass_.view()['fields'], view_json)

    def test_set_offer(self):
        self.assertEqual(len(self.pass_._create_payload()['fields']), 1)
        offer_data = {
            'multiUserOffer': True,
            'redemptionChannel': 'both',
            'provider': 'UA'
        }
        self.pass_.set_offer(
            multi_user_offer=True,
            redemption_channel='both',
            provider='UA'
        )
        self.assertEqual(
            self.pass_._google_features.top_level_fields['offerModule'],
            offer_data
        )
        json_payload = self.pass_._create_payload()
        self.assertEqual(len(json_payload['fields']), 2)
        self.assertEqual(
            json_payload['fields']['offerModule'],
            offer_data
        )

    def test_set_logo_image(self):
        self.pass_.set_logo_image(
            'https://urbanairship.com/cool_image.png',
            description='The logo image'
        )
        self.assertEqual(
            self.pass_.view()['top_level_fields']['image'],
            {
                'title.string': 'https://urbanairship.com/cool_image.png',
                'description.string': 'The logo image'
            }
        )
        self.assertEqual(
            self.pass_._create_payload()['fields']['image'],
            {
                'title.string': 'https://urbanairship.com/cool_image.png',
                'description.string': 'The logo image'
            }
        )

    def test_set_background_image(self):
        self.pass_.set_background_image(
            value='https://urbanairship.com/cool_image.png',
            description='The background image'
        )
        self.assertEqual(
            self.pass_.view()['top_level_fields'][ua.GoogleFieldType._IMAGE_MODULE],
            {
                'image': 'https://urbanairship.com/cool_image.png',
                'imageDescription': 'The background image'
            }
        )
        self.assertEqual(
            self.pass_._create_payload()['fields'][ua.GoogleFieldType._IMAGE_MODULE],
            {
                'image': 'https://urbanairship.com/cool_image.png',
                'imageDescription': 'The background image'
            }
        )

    def test_add_message(self):
        self.pass_.add_message(
            body='The message body',
            header='The message title',
            action_uri='https://urbanairship.com',
            action_uri_description='Cool website',
            image_uri='https://urbanairship.com/cool_image.png',
            image_description='Cool image',
            starttime=datetime.datetime(2016, 9, 15),
            endtime=datetime.datetime(2016, 10, 15),
        )

        # Test full example
        self.assertEqual(
            self.pass_._create_payload()['messages'][0],
            {
                'body': 'The message body',
                'header': 'The message title',
                'actionUri': 'https://urbanairship.com',
                'actionUriDescription': 'Cool website',
                'imageUri': 'https://urbanairship.com/cool_image.png',
                'imageDescription': 'Cool image',
                'startTime': '2016-09-15T00:00',
                'endTime': '2016-10-15T00:00'
            }
        )

        self.assertRaises(
            ValueError, self.pass_.add_message, header='A header, but no body!'
        )



import collections
import copy
import datetime
import logging

from six.moves.urllib import parse

from uareach import util


logger = logging.getLogger(__name__)

class FieldKey(util.Constant):
    KEY_CLASS = 'Field'

    CHANGE_MESSAGE = 'changeMessage'
    FORMAT_TYPE = 'formatType'
    HIDE_EMPTY = 'hideEmpty'
    ORDER = 'order'
    REQUIRED = 'required'
    FIELD_TYPE = 'fieldType'
    CURRENCY_CODE = 'currencyCode'
    TEXT_ALIGNMENT = 'textAlignment'
    LABEL = 'label'
    VALUE = 'value'

    # Specific to Google
    COL = 'col'
    ROW = 'row'

    # Specific to Apple
    NUMBER_STYLE = 'numberStyle'


class GoogleFieldType(util.Constant):
    INFO_MODULE = 'infoModuleData'
    TEXT_MODULE = 'textModulesData'
    LINKS_MODULE = 'linksModuleData'
    POINTS_MODULE = 'pointsModule'
    NOT_ASSIGNED = 'notAssigned'
    TITLE_MODULE = 'titleModule'
    ACCOUNT_MODULE = 'acctModule'

    _LOYALTY_POINTS = 'loyaltyPoints'
    _OFFER_MODULE = 'offerModule'
    _IMAGE_MODULE = 'imageModulesData'
    _MESSAGE_MODULE = 'messageModule'


class AppleFieldType(util.Constant):
    PRIMARY = 'primary'
    SECONDARY = 'secondary'
    TERTIARY = 'tertiary'
    AUXILIARY = 'auxiliary'
    BACK = 'back'
    HEADER = 'header'
    NOT_ASSIGNED = 'notAssigned'


class TextAlignment(util.Constant):
    LEFT = 'textAlignmentLeft'
    CENTER = 'textAlignmentCenter'
    RIGHT = 'textAlignmentRight'
    NATURAL = 'textAlignmentNatural'


class NumberStyle(util.Constant):
    DECIMAL = 'numberStyleDecimal'
    TEXT = 'numberStyleSpellOut'
    SCIENTIFIC = 'numberStyleScientific'
    PERCENT = 'numberStylePercent'


class CurrencyCode(util.Constant):
    USD = 'USD'
    EUR = 'EUR'
    CNY = 'CNY'
    JPY = 'JPY'
    GBP = 'GBP'
    RUB = 'RUB'
    AUD = 'AUD'
    CHF = 'CHF'
    CAD = 'CAD'
    HKD = 'HKD'
    SEK = 'SEK'
    NZD = 'NZD'
    KRW = 'KRW'
    SGD = 'SGD'
    NOK = 'NOK'
    MXN = 'MXN'
    INR = 'INR'


class FormatType(util.Constant):
    NUMBER = 'Number'
    STRING = 'String'
    URL = 'URL'
    DATE = 'Date'
    CURRENCY = 'Currency'
    PHONE = 'Phone'
    EMAIL = 'Email'


class Field(collections.MutableMapping):
    """Represent a Field object -- used by both AppleTemplate and
    GoogleTemplate to represent data/information on a template.
    """
    GOOGLE_LABEL_MAP = {
        GoogleFieldType.INFO_MODULE: 'label',
        GoogleFieldType.TEXT_MODULE: 'header',
        GoogleFieldType.LINKS_MODULE: 'description',
        GoogleFieldType._IMAGE_MODULE: 'imageDescription',
        GoogleFieldType.POINTS_MODULE: 'label',
        GoogleFieldType.NOT_ASSIGNED: 'label',
        GoogleFieldType.TITLE_MODULE: 'description.string',
        GoogleFieldType.ACCOUNT_MODULE: 'label',
        GoogleFieldType._LOYALTY_POINTS: 'label'
    }
    GOOGLE_VALUE_MAP = {
        GoogleFieldType.INFO_MODULE: 'value',
        GoogleFieldType.TEXT_MODULE: 'body',
        GoogleFieldType.LINKS_MODULE: 'uri',
        GoogleFieldType._IMAGE_MODULE: 'image',
        GoogleFieldType.POINTS_MODULE: 'balance',
        GoogleFieldType.NOT_ASSIGNED: 'value',
        GoogleFieldType.TITLE_MODULE: 'title.string',
        GoogleFieldType.ACCOUNT_MODULE: 'value',
        GoogleFieldType._LOYALTY_POINTS: 'balance'
    }

    def __init__(self, name=None, **kwargs):
        if not name:
            raise ValueError('Must set field name.')
        self.name = name
        self.field_values = {}
        for key, val in kwargs.iteritems():
            self[key] = val

    # Serialization / Deserialization methods for Apple & Google
    @classmethod
    def build_apple_field(cls, name, data):
        """Create a Field from an Apple-based field dictionary.

        Arguments:
            name (string): The field name (e.g. 'Reward Points')
            data (dict): The field dictionary.

        Raises:
            ValueError

        Returns:
            A Field object.

        Example:
            >>> Field.build_apple_field('Reward Points', {
            ...     'changeMessage': None,
            ...     'label': 'Points',
            ...     'hideEmpty': False,
            ...     'formatType': 'String',
            ...     'value': '1234.0',
            ...     'fieldType': 'header',
            ...     'required': False,
            ...     'order': 1
            ... }
            <Field name:Reward Points, fieldType:header>
        """
        AppleFieldType.validate(data.get('fieldType'))
        return cls(name, **data)

    @classmethod
    def build_google_field(cls, name, data):
        """Create a Field from a Google-based module dictionary.

        Arguments:
            name (string): The field name (e.g. 'Reward Points')
            data (dict): The field dictionary.

        Returns:
            A Field object.

        Example:
            >>> Field.build_google_field('Points', {
            ...     'col': 1,
            ...     'row': 0,
            ...     'label': 'Points',
            ...     'hideEmpty': False,
            ...     'formatType': 'String',
            ...     'balance': '1234.0',
            ...     'fieldType': 'pointsModuleData'
            ... }
            <Field name:Points, fieldType:header>
        """
        GoogleFieldType.validate(data.get('fieldType'), allow_private=True)
        field = cls(name, **data)

        # Post processing for google: convert row/col to order if possible.
        row, col = field.get('row'), field.get('col')
        row_col = [row, col]
        if 0 in row_col and not None in row_col:
            del field['row']
            del field['col']
            field['order'] = field._row_col_to_order(row, col)
        return field

    def build_apple_json(self):
        """Take a field object, fill in necessary blanks for serialization.

        Returns:
            A dictionary representing a field object for an Apple Template.
            Will attempt to infer formatType and numberStyle if not specified
            by the caller.

        Example:
            >>> f = Field(
            ...     name='Points',
            ...     label='Points',
            ...     value=1234,
            ...     fieldType=AppleFieldType.HEADER
            ... )
            >>> f.build_apple_json()
            {
                'label': 'Points',
                'value': 1234,
                'fieldType': 'header',
                'formatType': 'Number',
                'numberStyle': 'numberStyleDecimal'
            }
        """
        if self['fieldType'] == GoogleFieldType.NOT_ASSIGNED:
            return self.field_values

        payload = copy.deepcopy(self.field_values)
        if (
            payload.get('formatType') == 'Number' and
            not payload.get('numberStyle')
        ):
            payload['numberStyle'] = self._infer_number_style(self['value'])
        return payload

    def build_google_template_json(self):
        """This function converts a field object into a properly-formatted
        Google template field.

        Returns:
            A dictionary representing a field (module item) object for a Google
            Template. Will automatically convert value/label to the appropriate
            keys given the module type (e.g., header/body for TEXT_MODULE).

        Example:
            >>> f = Field(
            ...     name='Some Text',
            ...     label='Points',
            ...     value=1234,
            ...     fieldType=GoogleFieldType.TEXT_MODULE
            ... )
            >>> f.serialize_to_google()
            {'header': 'label', 'body': 1234, 'fieldType': 'textModule',
             'formatType': 'Number'}
        """
        if self['fieldType'] == GoogleFieldType.NOT_ASSIGNED:
            return self.field_values

        payload = {}
        for key, value in self.field_values.iteritems():
            if key == 'label':
                payload[self.GOOGLE_LABEL_MAP[self['fieldType']]] = value
            elif key == 'value':
                payload[self.GOOGLE_VALUE_MAP[self['fieldType']]] = value
            else:
                payload[key] = value
        payload = self._common_google_processing(payload)
        return payload

    def _build_common_template_json(self):
        """Common serialization items. Things that happen for both Apple
        and Google.

        .. note::

            * If currencyCode is specified and formatType isn't, validate
            currencyCode and set formatType to 'Currency'.
            * If formatType isn't specified, use the general _infer_format_type
            method to guess.
            * If no order is specified, default to 1.
            * If no fieldType is specified, default to ``notAssigned``.

        Returns:
            A dict object -- represents key val pairs not represented directly
            in the object but that can be inferred from the object data.
        """
        if self.get('currencyCode'):
            CurrencyCode.validate(self['currencyCode'])
            self['formatType'] = 'Currency' if not self.get('formatType') else self['formatType']
        if not self.get('formatType') and self.get('value'):
            self['formatType'] = self._infer_format_type(self['value'])
        if not self.get('order'):
            self['order'] = 1
        if not self.get('fieldType'):
            self['fieldType'] = 'notAssigned'

    def build_google_pass_json(self):
        """This function converts a field object into a properly-formatted
        Google pass field.

        Returns:
            A dictionary representing a field (module item) object for a Google
            Pass.

        Example:
            >>> f = Field(
            ...     name='Some Text',
            ...     label='Points',
            ...     value=1234,
            ...     fieldType=GoogleFieldType.TEXT_MODULE
            ... )
            >>> f.serialize_to_google()
            {'label': 'Points', 'value': 1234, 'fieldType': 'textModule'}
        """
        payload = copy.deepcopy(self.field_values)
        return self._common_google_processing(payload)

    def build_pass_json(self):
        """When creating a field for a pass, there may be no way to
        determine the fieldType. This method builds json for fields
        where the vendor/fieldType may be unknown.

        """
        try:
            # Handle known google fields
            GoogleFieldType.validate(self.get('fieldType'))
            return self.build_google_pass_json()
        except ValueError:
            pass

        try:
            AppleFieldType.validate(self.get('fieldType'))
            return self.build_apple_json()
        except ValueError:
            pass

        return self.field_values

    # Apple/Google Helper functions
    @staticmethod
    def _infer_number_style(value):
        """In the case of a number field, determine the number style. There's
        no real way to determine if numberStylePercent is the intended
        numberStyle.
        """
        if isinstance(value, float) or isinstance(value, int):
            return NumberStyle.DECIMAL
        elif isinstance(value, basestring):
            if 'E' in value or 'e' in value:
                return NumberStyle.SCIENTIFIC
            return NumberStyle.TEXT
        else:
            raise ValueError('Unrecognized number value: {}'.format(value))

    @staticmethod
    def _row_col_to_order(row, col):
        """Convert rows/cols (legacy) to order. (Google only)
        """
        if row == 0 and col == 0:
            return 1
        elif row == 0:
            return col + 1
        else:
            return row + 1

    @staticmethod
    def _infer_format_type(value):
        """Determine the format type of the field, if not declared
        explicitly. (Apple and Google).

        Notes:
            Should guess correctly in most cases, but formatType can always
            be overridden  by the user.
        """
        if isinstance(value, basestring):
            # Int/float values
            try:
                int(value)
                return 'Number'
            except ValueError:
                pass

            try:
                # Handles scientific notation as well.
                float(value)
                return 'Number'
            except ValueError:
                pass

            # Date values
            try:
                datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')
                return 'Date'
            except ValueError:
                pass

            # URL values
            url = parse.urlparse(value)
            if url.scheme != '' and url.netloc != '':
                return 'URL'

            return 'String'
        elif isinstance(value, float) or isinstance(value, int):
            return 'Number'
        else:
            raise ValueError('Unrecognized value type: {}'.format(type(value)))

    @staticmethod
    def _common_google_processing(payload):
        if (
            payload.get('order') and
            payload.get('fieldType') == GoogleFieldType.TEXT_MODULE
        ):
            # Special case: Order does not work in text modules -- must
            # reconvert to rows/cols.
            payload['col'], payload['row'] = 0, payload['order'] - 1
            del payload['order']
        return payload

    # Dictionary implementation
    def __setitem__(self, key, value):
        # Convert Google label/value pairs to standard label, value
        if key in Field.GOOGLE_LABEL_MAP.values():
            key = FieldKey.LABEL
        if key in Field.GOOGLE_VALUE_MAP.values():
            key = FieldKey.VALUE
        # Catch an invalid key
        FieldKey.validate(key)

        # Catch an invalid value
        if key == FieldKey.TEXT_ALIGNMENT:
            TextAlignment.validate(value)
            self.field_values[key] = value
        elif key == FieldKey.FORMAT_TYPE:
            FormatType.validate(value)
            self.field_values[key] = value
        elif key == FieldKey.NUMBER_STYLE:
            # Convert number style from PKNumberStyle to numberStyle, if necessary.
            value = value.replace('PKN', 'n')
            NumberStyle.validate(value)
            self.field_values[key] = value
        # Convert loyaltyPoints to pointsModule
        elif key == FieldKey.FIELD_TYPE:
            util.check_multiple(value, AppleFieldType, GoogleFieldType)
            if value == GoogleFieldType._LOYALTY_POINTS:
                self.field_values[key] = GoogleFieldType.POINTS_MODULE
            else:
                self.field_values[key] = value
        else:
            self.field_values[key] = value

    def __getitem__(self, key):
        return self.field_values[key]

    def __iter__(self):
        return iter(self.field_values)

    def __len__(self):
        return len(self.field_values)

    def __delitem__(self, key):
        del self.field_values[key]

    # Additional magic methods
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self.field_values == other.field_values and self.name == other.name

    def __repr__(self):
        return '<Field name:{}, fieldType:{}>'.format(
            self.name,
            self.get('fieldType')
        )



import logging
import json

import uareach as ua
from uareach import common, util
from uareach.templates import Template


logger = logging.getLogger(__name__)

class PassMetadata(util.Constant):
    PUBLIC_URL = 'publicUrl'
    PASS_ID = 'pass_id'
    TEMPLATE_ID = 'template_id'


def get_pass(reach, pass_id=None, external_id=None):
    """Retrieve a pass.

    Args:
        reach (reach.core.Reach): A uareach client object.
        pass_id (str): The pass ID of the pass you wish to retrieve.
        pass_external_id (str): The external ID of the pass you wish to
        retrieve.

    Returns:
        An ApplePass or GooglePass object.

    Example:
        >>> my_pass = Pass.get_pass(pass_external_id=12345)
        <id:67890745, templateId:51010>
    """
    if not (pass_id or external_id) or (pass_id and external_id):
        raise ValueError('Please specify only one of pass_id or external_id.')

    response = reach.request(
        method='GET',
        body=None,
        url=Pass.build_url(
            common.PASS_BASE_URL,
            main_id=pass_id,
            pass_external_id=external_id
        ),
        version=1.2
    )
    return _pass_dispatch(response.json())


def _pass_dispatch(data):
    """Infer the pass vendor based on field type.
    """
    fields = data.get('fields')
    for name, field_data in fields.iteritems():
        field_type = field_data['fieldType']
        if field_type == 'notAssigned':
            continue
        try:
            ua.AppleFieldType.validate(field_type)
            return ApplePass.from_data(data)
        except ValueError:
            try:
                ua.GoogleFieldType.validate(field_type)
                return GooglePass.from_data(data)
            except ValueError:
                pass
    raise ValueError("Unrecognized pass structure.")


def delete_pass(reach, pass_id=None, external_id=None):
    """Delete a pass.

    Arguments:
        reach (reach.Reach): A UA uareach object.
        pass_id (str or int): The ID of the pass you wish to delete.
        external_id (str or int): The external ID of the pass you wish
            to delete.

    Returns:
        A boolean representing success.

    Raises:
        ValueError: If neither, or both, of pass_id and external_id
            are specified.

    Example:
        >>> delete_pass(uareach, pass_id='123456')
        True
    """
    if not (pass_id or external_id) or (pass_id and external_id):
        raise ValueError('Please specify only one of pass_id or external_id.')

    response = reach.request(
        method='DELETE',
        body=None,
        url=Pass.build_url(
            common.PASS_BASE_URL,
            main_id=pass_id,
            pass_external_id=external_id
        ),
        version=1.2
    )
    logger.info('Successful pass deletion: {}'.format(
        pass_id if pass_id else external_id
    ))
    return True


def add_pass_locations(reach, locations, pass_id=None, external_id=None):
    """Add locations to a pass.

    Arguments:
        reach (reach.core.Reach): A uareach client object.
        locations (list): A list of dictionaries representing location
            objects.
        pass_id (str): The ID of the pass you wish to add
            locations to.
        external_id (str): The external ID of the pass you wish to add
            locations to.

    Returns:
        A list of objects containing the location value and passLocationId.

    Raises:
        ValueError: If neither, or both of, pass_id and external_id
            are specified.

    Example:
        >>> add_locations(ua_reach, location1, location2, pass_id=12345)
        True
    """
    if not (pass_id or external_id) or (pass_id and external_id):
        raise ValueError('Please specify only one of pass_id or external_id.')

    response = reach.request(
        method='POST',
        body=json.dumps({"locations": locations}),
        url=Pass.build_url(
            common.PASS_ADD_LOCATION_URL,
            main_id=pass_id,
            pass_external_id=external_id
        ),
        content_type='application/json',
        version=1.2
    )
    logger.info('Successfully added {} locations to pass {}.'.format(
        len(locations),
        pass_id if pass_id else external_id
    ))
    return response.json()


def delete_pass_location(reach, location_id, pass_id=None, external_id=None):
    """Delete a location from a pass.

    Arguments:
        reach (reach.core.Reach): A uareach client object.
        location_id (str or int): The ID of the location you wish to delete.
        pass_id (str or itn): The ID of the pass you wish to delete
            locations from.
        external_id (str or int): The external ID of the pass you wish
            to delete locations from.

    Returns:
        A boolean representing success.

    Raises:
        ValueError: If both or neither of pass_id and external_id are
            specified.

    Example:
        >>> delete_location(ua_reach, 12345, pass_id=44444)
        True
    """
    if not (pass_id or external_id) or (pass_id and external_id):
        raise ValueError('Please specify only one of pass_id or external_id.')

    response = reach.request(
        method='DELETE',
        body=None,
        url=Pass.build_url(
            common.PASS_DELETE_LOCATION_URL,
            main_id=pass_id,
            pass_external_id=external_id,
            location_id=location_id
        ),
        version=1.2
    )
    logger.info('Successfully deleted location {} from pass {}.'.format(
        location_id,
        pass_id if pass_id else external_id
    ))
    return True


class PassList(common.IteratorParent):
    """Forms a pass listing request.

    Arguments:
        reach (reach.core.Reach): The uareach client object.
        template_id (int): An integer ID specifing the template you would like to
            retrieve passes from.
        status (string): An optional string, can be one of ``"installed"``,
            ``"uninstalled"``, ``"been_installed"``, and ``"not_been_installed"``.
        page_size (int): The size of each page of the pass listing
            response. Defaults to 10.
        page (int): The page to start the listing response on. Defaults
            to 1.
        order (string): The attribute used to order the passes. Can be one
            of ``"id"``, ``"name"``, ``"createdAt"``, or ``"updatedAt"``.
            Defaults to ``"id"``.
        direction (string): The direction of the pass list, can be one of
            ``"ASC"`` or ``"DESC"``. Defaults to ``"DESC"``.

    Example:
        >>> pass_list = PassList(ua_reach)
        >>> for item in pass_list:
        ...     print item['createdAt']
        ...
        2016-07-27T21:14:47Z
        2016-07-23T12:13:32Z
        2016-06-09T17:32:14Z
    """
    base_url = common.PASS_BASE_URL.format('')
    data_attribute = 'passes'

    def __init__(
        self,
        reach,
        status=None,
        template_id=None,
        page_size=None,
        page=None,
        order=None,
        direction=None
    ):

        if direction not in {None, 'ASC', 'DESC'}:
            raise ValueError(
                'Unrecognized direction "{}". '
                'Please use one of "ASC" or "DESC"'.format(direction)
            )

        statuses = {
            'installed',
            'uninstalled',
            'been_installed',
            'not_been_installed'
        }

        if not (status is None or status in statuses):
            raise ValueError(
                "Unrecognized status '{}'. "
                "Please use one of {}.".format(status, statuses)
            )

        params = {
            'status': status,
            'templateId': template_id,
            'pageSize': page_size,
            'page': page,
            'order': order,
            'direction': direction
        }

        params = {k: v for k, v in params.iteritems() if v is not None}
        super(PassList, self).__init__(reach, params)


@util.inherit_docs
class Pass(Template):

    METADATA = ['publicUrl']
    READ_ONLY_METADATA = [
        'url', 'status', 'updatedAt', 'pass_id', 'template_id', 'createdAt', 'serialNumber', 'tags'
    ]

    def create(self, reach, template_id=None, template_ext_id=None, pass_ext_id=None):
        """Create a pass.

        Arguments:
            reach (reach.core.Reach): A uareach client object.
            template_id (str/int): A template ID.
            template_ext_id (str/int): A template external ID.
            pass_ext_id (str/int): A pass external ID.

        Returns:
            A dictionary containing ``'id'`` and ``'url'`` keys.

        Raises:
            ValueError: If neither, or both, of template_id and template_ext_id
                are specified.

        Example:
            >>> my_pass.create(ua_reach, project_id=12345)
            {'id': 123456, 'url': 'https://blah.com'}
        """
        if template_id and template_ext_id:
            raise ValueError(
                "'template_id' and 'template_ext_id' cannot both be set"
            )

        response = reach.request(
            method='POST',
            body=json.dumps(self._create_payload()),
            url=Pass.build_url(
                common.PASS_BASE_URL,
                main_id=template_id,
                template_external_id=template_ext_id,
                pass_external_id=pass_ext_id
            ),
            content_type='application/json',
            version=1.2
        )
        pass_id = response.json()['id']
        logger.info('Successful pass creation: {}'.format(
            pass_id if pass_id else pass_ext_id
        ))
        self.metadata['id'] = pass_id
        return {'id': pass_id, 'url': response.json()['url']}

    def update(self, reach, pass_id=None, external_id=None):
        """Update a pass.

        Arguments:
            reach (reach.core.Reach): A uareach client object.
            pass_id (str): ID of the pass to update.
            external_id (str): External ID of the pass to update.

        Returns:
            A dictionary containing a ticketId.

        Raises:
            ValueError: If neither, or both, of pass_id and external_id
                are specified.

        Example:
            >>> my_pass.update(ua_reach)
            {'ticketId': 4125123}
        """
        if pass_id and external_id:
            raise ValueError("'pass_id' and 'external_id' cannot both be set")

        if not (pass_id or external_id):
            pass_id = self.metadata.get('id', None)
            if not pass_id:
                raise ValueError(
                    "Either set pass_id or external_id when "
                    "calling update, or set the pass 'id' attribute."
                )

        response = reach.request(
            method='PUT',
            body=json.dumps(self._create_payload()),
            url=Pass.build_url(
                common.PASS_BASE_URL,
                main_id=pass_id,
                template_external_id=external_id
            ),
            content_type='application/json',
            version=1.2
        )
        logger.info('Successful pass update: {}'.format(
            pass_id if pass_id else external_id
        ))
        return response.json()

    @classmethod
    def from_data(cls, data):
        pass_ = cls()

        # Handle metadata
        non_metadata_keys = {'fields', 'headers', 'beacons', 'messages'}
        pass_.metadata = {}
        for key in data.keys():
            if key not in non_metadata_keys:
                pass_.metadata[key] = data.pop(key)

        if pass_.metadata.get('templateId'):
            pass_.metadata[PassMetadata.TEMPLATE_ID] = pass_.metadata['templateId']
            del pass_.metadata['templateId']
        if pass_.metadata.get('id'):
            pass_.metadata[PassMetadata.PASS_ID] = pass_.metadata['id']
            del pass_.metadata['id']

        # Handle headers
        pass_.headers = data.pop('headers', {})
        if pass_.headers.get(ua.TemplateHeader.BARCODE_TYPE):
            no_pk = pass_.headers[ua.TemplateHeader.BARCODE_TYPE]['value'].replace('PKB', 'B')
            pass_.headers[ua.TemplateHeader.BARCODE_TYPE]['value'] = no_pk

        return pass_

    def view(self):
        payload = super(Pass, self)._create_payload()
        payload['fields'] = {}
        for name, field in self.fields.iteritems():
            payload['fields'][name] = field.build_pass_json()
        for name, item in self.metadata.iteritems():
            payload[name] = item
        return {key: val for key, val in payload.iteritems() if val}

    def _create_payload(self):
        payload = self.view()
        payload.get('publicUrl', {}).pop('path', None)
        payload.get('publicUrl', {}).pop('image', None)
        return {key: val for key, val in payload.iteritems() if
                key not in self.READ_ONLY_METADATA}

    def _add_metadata_helper(self, name, value):
        # publicUrl metadata is handled differently from other metadata keys
        if name == PassMetadata.PUBLIC_URL:
            self.set_public_url(value)
        else:
            super(Pass, self)._add_metadata_helper(name, value)

    # Additional convenience methods
    def set_expiration(self, date):
        """Set the pass expiration date.

        Arguments:
            date (datetime): A python datetime object.

        Example:
            >>> my_pass.set_expiration(datetime.datetime(2016,10,13))
        """
        # Implementation differs between apple/google. See overrides for
        # details.
        raise NotImplementedError

    def set_public_url(self, type_):
        """Set the public URL type.

        Arguments:
            type_ (string): A string representing the public URL type. Can be one of
                "single" or "multiple".

        Raises:
            ValueError: If ``type_`` is not one of "single" or "multiple".

        Example:
            >>> my_pass.set_public_url('single')
        """
        if type_ not in {'single', 'multiple'}:
            raise ValueError(
                "Unrecognized publicUrl type '{}'. Please use one of 'single' "
                "or 'multiple'.".format(type_)
            )
        self.metadata['publicUrl'] = {'type': type_}

    @staticmethod
    def build_url(
        base_url,
        main_id=None,
        template_external_id=None,
        pass_external_id=None,
        location_id=None
    ):
        if base_url == common.PASS_ADD_LOCATION_URL and main_id:
            return base_url.format(main_id)
        elif base_url == common.PASS_ADD_LOCATION_URL:
            return base_url.format('id/' + str(pass_external_id))
        elif base_url == common.PASS_DELETE_LOCATION_URL and main_id:
            return base_url.format(main_id, location_id)
        elif base_url == common.PASS_DELETE_LOCATION_URL:
            return base_url.format('id/' + str(pass_external_id), location_id)
        elif main_id and not (template_external_id or pass_external_id):
            return base_url.format(main_id)
        elif template_external_id and not (main_id or pass_external_id):
            return base_url.format('id/' + str(external_id))
        elif pass_external_id and main_id and not template_external_id:
            return base_url.format(str(main_id) + '/id/' + str(pass_external_id))
        elif template_external_id and pass_external_id and not main_id:
            return base_url.format(
                'id/' + str(template_external_id) + '/id/' + str(pass_external_id)
            )
        else:
            return base_url.format('')

    def __repr__(self):
        """Note: Cannot gaurantee that any displayable attributes will be
        present, so we revert back to the default __repr__ implementation.
        """
        return '<{}.{} object at {}>'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self))
        )


@util.inherit_docs
class ApplePass(Pass):

    def __init__(self):
        super(ApplePass, self).__init__()
        self._apple_features = ua.AppleTemplate()

    @classmethod
    def from_data(cls, data):
        pass_ = super(ApplePass, cls).from_data(data)
        pass_._apple_features.beacons = data.pop('beacons', [])
        fields = data.pop('fields', {})
        for name, json_field in fields.iteritems():
            pass_.fields[name] = ua.Field.build_apple_field(
                name, json_field
            )
        return pass_

    def view(self):
        payload = super(ApplePass, self).view()
        payload['beacons'] = self._apple_features.beacons
        payload.update(self.metadata)
        return {key: val for key, val in payload.iteritems() if val}

    def set_expiration(self, date):
        date_str = date.strftime('%Y-%m-%dT%H:%M')
        self.headers['expirationDate'] = {'value': date_str}

    @util.set_docstring(ua.AppleTemplate.add_beacon)
    def add_beacon(self, uuid, relevant_text=None, major=None, minor=None):
        self._apple_features.add_beacon(uuid, relevant_text, major, minor)

    @util.set_docstring(ua.AppleTemplate.remove_beacon)
    def remove_beacon(self, uuid):
        self._apple_features.remove_beacon(uuid)

    @util.set_docstring(ua.AppleTemplate.set_logo_image)
    def set_logo_image(self, value):
        self.add_headers(logo_image=value)


@util.inherit_docs
class GooglePass(Pass):

    def __init__(self):
        super(GooglePass, self).__init__()
        self._google_features = ua.GoogleTemplate()

    @classmethod
    def from_data(cls, data):
        pass_ = super(GooglePass, cls).from_data(data)
        for name, field in data.pop('fields', {}).iteritems():
            if name == 'image':
                pass_._google_features.top_level_fields['image'] = field
            elif name == ua.GoogleFieldType._IMAGE_MODULE:
                pass_._google_features.add_top_level_fields(
                    ua.GoogleFieldType._IMAGE_MODULE,
                    **field
                )
            elif name == ua.GoogleFieldType._MESSAGE_MODULE:
                pass_._google_features.messages = field
            elif name == ua.GoogleFieldType._OFFER_MODULE:
                pass_._google_features.add_top_level_fields(
                    ua.GoogleFieldType._OFFER_MODULE,
                    **field
                )
            else:
                pass_.fields[name] = ua.Field.build_google_field(name, field)
        return pass_

    def view(self):
        payload = super(GooglePass, self).view()
        if self._google_features.top_level_fields:
            payload['top_level_fields'] = self._google_features.top_level_fields
        if self._google_features.messages:
            payload['messages'] = self._google_features.messages
        return payload

    def _create_payload(self):
        payload = super(GooglePass, self)._create_payload()
        if payload.get('top_level_fields'):
            additional_fields = payload.pop('top_level_fields')
            if not payload.get('fields'):
                payload['fields'] = {}
            payload['fields'].update(additional_fields)
        return payload

    def set_expiration(self, date):
        date_str = date.strftime('%Y-%m-%dT%H:%M')
        self._google_features.top_level_fields['endTime'] = {'value': date_str}

    # Google-specific methods
    @util.set_docstring(ua.GoogleTemplate.set_logo_image)
    def set_logo_image(self, value, description=None):
        if not description:
            description = ''

        self._google_features.top_level_fields['image'] = {
            'title.string': value,
            'description.string': description
        }

    @util.set_docstring(ua.GoogleTemplate.set_background_image)
    def set_background_image(self, value, description=None):
        self._google_features.set_background_image(value, description)

    @util.set_docstring(ua.GoogleTemplate.set_offer)
    def set_offer(
        self,
        multi_user_offer=None,
        endtime=None,
        provider=None,
        redemption_channel=None
    ):
        self._google_features.set_offer(
            multi_user_offer, endtime, provider, redemption_channel
        )

    @util.set_docstring(ua.GoogleTemplate.add_message)
    def add_message(
        self,
        body=None,
        header=None,
        action_uri=None,
        action_uri_description=None,
        image_uri=None,
        image_description=None,
        starttime=None,
        endtime=None
    ):
        self._google_features.add_message(
            body,
            header,
            action_uri,
            action_uri_description,
            image_uri,
            image_description,
            starttime,
            endtime
        )



from .core import Reach
from .common import Unauthorized, ReachFailure
from .templates import (
    TemplateMetadata,
    TemplateType,
    ProjectType,
    Type,
    TemplateHeader,
    BarcodeType,
    TransitType,
    get_template,
    delete_template,
    duplicate_template,
    add_template_locations,
    delete_template_location,
    TemplateList,
    AppleTemplate,
    GoogleTemplate,
    TemplateList,
)

from .fields import (
    FieldKey,
    GoogleFieldType,
    AppleFieldType,
    NumberStyle,
    TextAlignment,
    CurrencyCode,
    Field,
)

from .util import (
    rgb,
)

from .passes import (
    add_pass_locations,
    delete_pass_location,
    delete_pass,
    get_pass,
    PassList,
    ApplePass,
    GooglePass,
)



import unittest
import uareach as ua
from uareach.fields import Field


class TestField(unittest.TestCase):
    def setUp(self):
        self.min_apple_json = ('Member Name', {
            'fieldType': 'primary',
            'label': 'Member Name',
            'value': 'First Last'
        })
        self.min_apple_field = Field(
            name='Member Name',
            label='Member Name',
            value='First Last',
            fieldType=ua.AppleFieldType.PRIMARY
        )
        self.min_google_json = ('Member Name', {
            'label': 'Member Name',
            'value': 'First Last',
            'fieldType': ua.GoogleFieldType.INFO_MODULE
        })
        self.min_google_field = Field(
            name='Member Name',
            label='Member Name',
            value='First Last',
            fieldType=ua.GoogleFieldType.INFO_MODULE
        )

        self.full_apple_json = ('Member Name', {
            'fieldType': 'primary',
            'changeMessage': 'Things change',
            'formatType': 'String',
            'hideEmpty': True,
            'label': 'Member Name',
            'order': 2,
            'required': False,
            'value': 'First Last',
            'textAlignment': 'textAlignmentRight'
        })
        self.full_apple_field = Field(
            name='Member Name',
            fieldType=ua.AppleFieldType.PRIMARY,
            changeMessage='Things change',
            formatType='String',
            hideEmpty=True,
            label='Member Name',
            order=2,
            required=False,
            value='First Last',
            textAlignment='textAlignmentRight'
        )
        self.full_google_json = ('Member Name', {
            'order': 0,
            'label': 'Member Name',
            'value': 'Taylor Lakehurst',
            'fieldType': ua.GoogleFieldType.ACCOUNT_MODULE,
            'formatType': 'String',
            'changeMessage': None,
            'hideEmpty': False,
            'required': False
        })
        self.full_google_field = Field(
            order=0,
            name='Member Name',
            label='Member Name',
            value='Taylor Lakehurst',
            fieldType=ua.GoogleFieldType.ACCOUNT_MODULE,
            formatType='String',
            changeMessage=None,
            hideEmpty=False,
            required=False
        )

    def test_build_apple_field(self):
        self.assertEqual(
            Field.build_apple_field(
                self.min_apple_json[0], self.min_apple_json[1]
            ), self.min_apple_field
        )
        self.assertEqual(
            Field.build_apple_field(
                self.full_apple_json[0], self.full_apple_json[1]
            ), self.full_apple_field
        )

    def test_build_google_field(self):
        self.assertEqual(
            Field.build_google_field(
                self.min_google_json[0], self.min_google_json[1]
            ), self.min_google_field
        )
        self.assertEqual(
            Field.build_google_field(
                self.full_google_json[0], self.full_google_json[1]
            ), self.full_google_field
        )

    def test_build_apple_json(self):
        extended_apple_json = self.min_apple_json[1].copy()
        extended_apple_json.update({
            'order': 1,
            'formatType': 'String'
        })

        self.min_apple_field._build_common_template_json()
        self.assertEqual(
            extended_apple_json, self.min_apple_field.build_apple_json()
        )
        self.assertEqual(
            self.full_apple_json[1], self.full_apple_field.build_apple_json()
        )

    def test_build_google_json(self):
        extended_google_json = self.min_google_json[1].copy()
        extended_google_json.update({
            'order': 1,
            'formatType': 'String'
        })

        self.min_google_field._build_common_template_json()
        self.assertEqual(
            extended_google_json, self.min_google_field.build_google_template_json()
        )
        self.assertEqual(
            self.full_google_json[1], self.full_google_field.build_google_template_json()
        )

    def test_build_generic_json(self):
        field = Field(name='Member Name', value='First Last')
        payload = field.build_pass_json()
        self.assertEquals(payload, {'value': 'First Last'})

        google_field = Field(
            name='Test', fieldType=ua.GoogleFieldType.TEXT_MODULE, value='Hello'
        )
        payload = google_field.build_pass_json()
        self.assertEquals(payload, {'fieldType': 'textModulesData', 'value': 'Hello'})

        apple_field = Field(
            name='Test', fieldType=ua.AppleFieldType.PRIMARY, formatType='Number', value=1234
        )
        payload = apple_field.build_pass_json()
        self.assertEquals(
            payload,
            {
                'formatType': 'Number',
                'value': 1234,
                'fieldType': 'primary',
                'numberStyle': 'numberStyleDecimal'
            }
        )

    def test_build_common_json(self):
        # Test setting Currency formatType
        field = Field(name='Value', currencyCode='USD')
        field._build_common_template_json()
        self.assertEquals(field.field_values['formatType'], 'Currency')

        # Test that infer_format_type is triggered
        field = Field(name='Member Name', value='Hello there')
        field._build_common_template_json()
        self.assertEquals(field.field_values['formatType'], 'String')

        # Test that order and fieldType are set
        field = Field(name='Member Name')
        field._build_common_template_json()
        self.assertEquals(field.field_values['order'], 1)
        self.assertEquals(field.field_values['fieldType'], 'notAssigned')

    def test_infer_number_style(self):
        self.assertEqual(
            Field._infer_number_style(1234), ua.NumberStyle.DECIMAL
        )
        self.assertEqual(
            Field._infer_number_style(1234.0142), ua.NumberStyle.DECIMAL
        )
        self.assertEqual(
            Field._infer_number_style('1234'), ua.NumberStyle.TEXT
        )
        self.assertEqual(
            Field._infer_number_style('1234e2'), ua.NumberStyle.SCIENTIFIC
        )
        self.assertEqual(
            Field._infer_number_style('1234E2'), ua.NumberStyle.SCIENTIFIC
        )
        self.assertRaises(ValueError, Field._infer_number_style, {})

    def test_row_col_to_order(self):
        self.assertEqual(Field._row_col_to_order(0, 0), 1)
        self.assertEqual(Field._row_col_to_order(1, 0), 2)
        self.assertEqual(Field._row_col_to_order(0, 1), 2)
        row_col_data = {
            'row': 0,
            'col': 1,
            'label': 'Member Name',
            'value': 'Taylor Lakehurst',
            'fieldType': 'acctModule',
            'formatType': 'String'
        }
        row_col_field = Field.build_google_field('Member Name', row_col_data)
        self.assertEqual(row_col_field['order'], 2)
        self.assertIsNone(row_col_field.get('row', None))
        self.assertIsNone(row_col_field.get('col', None))

    def test_infer_format_type(self):
        self.assertEqual(Field._infer_format_type('Hello'), 'String')
        self.assertEqual(Field._infer_format_type(1234), 'Number')
        self.assertEqual(Field._infer_format_type(123.321), 'Number')
        self.assertEqual(Field._infer_format_type('123.01'), 'Number')
        self.assertEqual(Field._infer_format_type('123'), 'Number')
        self.assertEqual(Field._infer_format_type('1.233e4'), 'Number')
        self.assertEqual(
            Field._infer_format_type('http://google.com/cool_gif.gif'), 'URL'
        )
        self.assertEqual(
            Field._infer_format_type('2016-01-01T19:00:00Z'), 'Date'
        )
        self.assertRaises(ValueError, Field._infer_format_type, {})

    def test_setitem(self):
        gf = Field(
            name='Member Name',
            fieldType=ua.GoogleFieldType.TEXT_MODULE
        )
        af = Field(
            name='Member Name',
            fieldType=ua.AppleFieldType.PRIMARY
        )
        # Test google field label/value conversions
        gf['header'] = 'Hello hello'
        gf['body'] = 'This is the body'
        self.assertEquals(gf['label'], 'Hello hello')
        self.assertEquals(gf['value'], 'This is the body')
        # Test invalid key
        self.assertRaises(ValueError, gf.__setitem__, 'doesntExist', 'Who Cares?')
        # Test invalid values
        self.assertRaises(ValueError, af.__setitem__, 'textAlignment', 'blah')
        self.assertRaises(ValueError, af.__setitem__, 'formatType', 'bool')
        self.assertRaises(ValueError, af.__setitem__, 'numberStyle', 'wrong')
        # Test numberStyle replacement
        af['numberStyle'] = 'PKNumberStyleDecimal'
        self.assertEquals(af['numberStyle'], 'numberStyleDecimal')
        # Test loyaltyPoints replacement
        gf['fieldType'] = ua.GoogleFieldType._LOYALTY_POINTS
        self.assertEquals(gf['fieldType'], ua.GoogleFieldType.POINTS_MODULE)

    def test_other_field_manipulations(self):
        f = Field(
            name='Member Name',
            label='Member Name',
            value='First Last',
            fieldType=ua.AppleFieldType.PRIMARY
        )
        f['label'] = 'Updated Member Name'
        f['formatType'] = 'String'
        updated_f = Field(
            name='Member Name',
            label='Updated Member Name',
            value='First Last',
            formatType='String',
            fieldType=ua.AppleFieldType.PRIMARY
        )
        self.assertEqual(f, updated_f)
        del f['formatType']
        updated_f = Field(
            name='Member Name',
            label='Updated Member Name',
            value='First Last',
            fieldType=ua.AppleFieldType.PRIMARY
        )
        self.assertEqual(f, updated_f)
        self.assertEqual(len(f), 3)




import inspect
import types

import uareach as ua


class Constant(object):
    KEY_CLASS=None

    @classmethod
    def validate(cls, item, allow_private=False):
        if allow_private:
            underscores = '__'
        else:
            underscores = '_'

        class_dict = {
            (cls.__name__ + '.' + attr): val for
            attr, val in cls.__dict__.iteritems() if
            not callable(attr) and not attr.startswith(underscores)
            and not attr == 'KEY_CLASS'
        }

        if item not in class_dict.values():
            raise ValueError(
                "Unrecognized option '{}'. {}".format(
                    item,
                    cls._build_error_string(class_dict)
                )
             )

    @classmethod
    def _build_error_string(cls, class_dict):
            col1 = max([len(attr) for attr in class_dict.keys()])
            col2 = max([len(str(val)) for val in class_dict.values()])
            table = ""
            if not cls.KEY_CLASS:
                row_format = '{:<' + str(col1) + '}  {:<' + str(col2) + '}\n'
                table += row_format.format('Enum', 'Corresponding String')
                table += row_format.format(
                    '-' * max(col1, 4),
                    '-' * max(col2, 20)
                )
                for key, val in class_dict.iteritems():
                    table += row_format.format(key, val)
                return "Valid options for '{}' are:\n\n{}".format(cls.__name__, table)
            row_format = '{:<' + str(col1) + '}\n'
            table += row_format.format('Keys')
            table += row_format.format('-' * max(col2, 4))
            for val in class_dict.values():
                table += row_format.format(val)
            return "Valid keys for '{}' are:\n\n{}".format(cls.KEY_CLASS, table)


def check_multiple(item, *args):
    errors = []
    for enum in args:
        try:
            enum.validate(item, allow_private=True)
        except ValueError as e:
            errors.append(str(e))
    if len(errors) == len(args):
        raise ValueError('\n'.join(errors))


def inherit_docs(cls):
    """A class decorator that will pull method docstrings from a parent class
    if not specified in the child class.
    """
    for name, func in vars(cls).items():
        class_method = False
        if isinstance(func, classmethod):
            func = func.__func__
            class_method = True

        if isinstance(func, types.FunctionType) and not func.__doc__:
            for parent in inspect.getmro(cls):
                parfunc = getattr(parent, name, None)
                if class_method:
                    parfunc = parfunc.__func__
                if parfunc and getattr(parfunc, '__doc__', None):
                    func.__doc__ = parfunc.__doc__
                    break
    return cls


def set_docstring(method):
    def _decorator(func):
        func.__doc__ = method.__doc__
        return func
    return _decorator


def rgb(red, green, blue):
    """Helper function that returns rgb strings to be used
    by color-specification headers.

    Arguments:
        red (int): Red color value.
        green (int): Green color value.
        blue (int): Blue color value.

    Returns:
        A string representing the specified color.

    Example:
        >>> rgb(23, 167, 42)
        'rgb(23, 167, 42)'
    """
    return 'rgb({0:d},{1:d},{2:d})'.format(red, green, blue)



import unittest

from uareach import util


class TestEnum(unittest.TestCase):
    def test_validate(self):
        class ExampleEnum(util.Constant):
            KEY1 = 'value1'
            KEY2 = 'value2'
            _PRIVATE_KEY = 'privatevalue'

        ExampleEnum.validate('value1')
        ExampleEnum.validate('value2')
        self.assertRaises(ValueError, ExampleEnum.validate, 'value3')
        self.assertRaises(ValueError, ExampleEnum.validate, 'privatevalue')


def example_function():
    """Here are some docs"""
    return True


@util.set_docstring(example_function)
def example_function_no_docs():
    return False


class WithDocs(object):

    def __init__(self):
        pass

    def method_1(self):
        """method 1 docs"""
        return True

    def method_2(self):
        """method 2 docs"""
        return True

    def method_3(self):
        """method 3 docs"""
        return True


@util.inherit_docs
class WithoutDocs(WithDocs):

    def __init__(self):
        pass

    def method_1(self):
        return True

    def method_2(self):
        return True

    def method_3(self):
        """this method 3 is different"""
        return False


class TestDocs(unittest.TestCase):

    def test_set_docstring(self):
        self.assertEquals(
            example_function.__doc__, example_function_no_docs.__doc__
        )

    def test_inherit_docs(self):
        self.assertEquals(
            WithDocs.method_1.__doc__, WithoutDocs.method_1.__doc__
        )

        self.assertEqual(
            WithDocs.method_2.__doc__, WithoutDocs.method_2.__doc__
        )

        self.assertEquals(
            WithoutDocs.method_3.__doc__, "this method 3 is different"
        )



import logging
import requests

from uareach import common, __about__


logger = logging.getLogger(__name__)


class Reach(object):

    def __init__(self, email_addy, api_key):
        self.email_addy = email_addy
        self.api_key = api_key

        self.session = requests.Session()
        self.session.auth = (email_addy, api_key)

    def request(self, method, body, url, content_type=None, version=None,
                params=None):
        return self._request(method, body, url, content_type, version, params)

    def _request(self, method, body, url, content_type=None, version=None,
                 params=None, encoding=None):
        headers = {
            'User-agent': 'UAReachPythonLib/{0}'.format(
                __about__.__version__
            )
        }

        if content_type:
            headers['Content-type'] = content_type
        if version:
            headers['Api-Revision'] = str(version)
        if encoding:
            headers['Content-Encoding'] = encoding

        logger.debug(
             'Making %s request to %s. Headers:\n\t%s\nBody:\n\t%s', method,
             url, '\n\t'.join(
                '%s: %s' % (key, value) for (key, value) in headers.items()
             ),
             body
        )

        response = self.session.request(
            method, url, data=body, params=params, headers=headers
        )

        logger.debug(
            'Received %s response. Headers:\n\t%s\nBody:\n\t%s',
            response.status_code,
            '\n\t'.join(
                '%s: %s' % (key, value) for (key, value)
                in response.headers.items()
            ),
            response.content
        )

        if response.status_code == 401:
            raise common.Unauthorized
        elif not (200 <= response.status_code < 300):
            raise common.ReachFailure.from_response(response)

        return response



import logging
import json

import six


BASE_URL = 'https://wallet-api.urbanairship.com/v1'

# Template URLs
TEMPLATE_BASE_URL = BASE_URL + '/template/{0}'
TEMPLATE_DUPLICATE_URL = BASE_URL + '/template/duplicate/{0}'
TEMPLATE_ADD_LOCATION_URL = BASE_URL + '/template/{0}/locations/'
TEMPLATE_REMOVE_LOCATION_URL = BASE_URL + '/template/{0}/location/{1}'

# Pass URLs
PASS_BASE_URL = BASE_URL + '/pass/{0}'
PASS_ADD_LOCATION_URL = BASE_URL + '/pass/{0}/locations'
PASS_DELETE_LOCATION_URL = BASE_URL + '/pass/{0}/location/{1}'


logger = logging.getLogger(__name__)


class Unauthorized(Exception):
    """Raised when we get a 401 from the server"""


class ReachFailure(Exception):
    """Raised when we get an error response from the server."""
    def __init__(self, error, error_code, details, response, *args):
        self.error = error
        self.error_code = error_code
        self.details = details
        self.response = response
        super(ReachFailure, self).__init__(*args)

    @classmethod
    def from_response(cls, response):
        """Instantiate a ValidationFailure from a Response object

        Arguments:
            response (response object): Response object used to create a
                failure object.

        Returns:
            A ReachFailure object.
        """
        try:
            payload = response.json()
            error = payload.get('error')
            error_code = payload.get('error_code')
            details = payload.get('details')
        except ValueError:
            error = response.reason
            error_code = None
            details = response.content

        logger.error(
            "Request failed with status %d: '%s %s': %s",
            response.status_code, error_code, error, json.dumps(details))

        return cls(
            error,
            error_code,
            details,
            response,
            response.status_code,
            response.content
        )


class IteratorParent(six.Iterator):
    """An iterator class for listing responses."""
    base_url = None
    data_attribute = None

    def __init__(self, reach, params):
        self.reach = reach
        self.params = params
        self._token_iter = iter(())
        self.page_count = None
        self._page = None

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return next(self._token_iter)
        except StopIteration:
            if self._load_page():
                return next(self._token_iter)
            else:
                raise StopIteration

    def _load_page(self):
        if not self.base_url:
            raise ValueError('base_url cannot be None.')

        response = self.reach.request(
            method='GET',
            body=None,
            url=self.base_url,
            version=1.2,
            params=self.params
        )
        self._page = response.json()
        self.params = self._page.get('pagination')
        if not self.page_count:
            self.page_count = self._page['count']/self.params['pageSize'] + 1
        if self.params['page'] > self.page_count:
            return False
        self.params['page'] += 1
        self._token_iter = iter(self._page[self.data_attribute])
        return True



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

__about__ = {}

with open("uareach/__about__.py") as fp:
    exec(fp.read(), None, __about__)

setup(
    name="uareach",
    version=__about__["__version__"],
    author="Urban Airship",
    author_email="support@urbanairship.com",
    url="http://urbanairship.com/",
    description="Python package for using the Urban Airship Reach API",
    long_description=open('README.rst').read(),
    packages=["uareach"],
    license='BSD License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries'
    ],
    install_requires=[
        'requests>=2.7',
        'six'
    ],
)
